d3.custom.chart.flow = function() {

    // public variables with default settings
    var //margin = {top:10, right:10, bottom:10, left:10}, // defaults
        margin = {top:15, right:15, bottom:15, left:15}, // defaults
        //padding = {top:20, right:10, bottom:10, left:10},
        padding = {top:25, right:15, bottom:15, left:15},
        transitionDuration = 300,
        chartGroup,
        container,
        svg,
        width,
        height,
        root,
        rootDomNode;

    var flow = d3.custom.layout.flow()
        .margin(margin)
        .padding(padding)
        .nodeWidth(110)
        .nodeHeight(30)
        .containerHeight(20);
     
    function chart(selection) {
        rootDomNode = selection.node();

        function debounce(fn, timeout) {
            var timeoutID = -1;
            return function() {
                if (timeoutID > -1) {
                    window.clearTimeout(timeoutID);
                }
                timeoutID = window.setTimeout(fn, timeout);
            }
        }

        function resize(selectedNode) {
            var flowWidth = root.width;
            flow.width( flowWidth );

            chart.update(selectedNode);
            
            svg.transition().duration(transitionDuration)
                .attr("width", function(d) {
                    return d.width + margin.left + margin.right;
                })
                .attr("height", function(d) {
                    return d.height + margin.top + margin.bottom;
                })
                .select(".chartGroup")
                .attr("width", function(d) {
                    return flowWidth;
                })
                .attr("height", function(d) {
                    return d.height + margin.top + margin.bottom;
                })
                .select(".background")
                .attr("width", function(d) {
                    return flowWidth;
                })
                .attr("height", function(d) {
                    return d.height + margin.top + margin.bottom;
                });
        }


//        d3.select(window).on('resize', function() {
//            debounce(resize, 50)();
//        });


        selection.each(function(arg) {
            root = arg;
            container = d3.select(this);

            var i = 0;

            if (!svg) {
                svg = container.append("svg")
                    .attr("class", "svg chartSVG")
                    .attr("transform", "translate(0, 0)")
                    .style("shape-rendering", "auto") // shapeRendering options; [crispEdges|geometricPrecision|optimizeSpeed|auto]
                    .style("text-rendering", "auto"); // textRendering options;  [auto|optimizeSpeed|optimizeLegibility|geometricPrecision]
                chartGroup = svg.append("svg:g")
                    .attr("class", "chartGroup");
                chartGroup.append("svg:rect")
                    .attr("class", "background");
                
                // Build the arrow marker for edges
                svg.append( "svg:defs" ).selectAll( "marker" )
                  .data( ["end"] )
                  .enter()
                  .append( "svg:marker" )
                  .attr( "id", String )
                  .attr( "viewBox", "0 -5 10 10" )
                  .attr( "refX", 6 )
                  .attr( "refY", 0 )
                  .attr( "markerWidth", 6 )
                  .attr( "markerHeight", 6 )
                  .attr( "orient", "auto" )
                  .append( "svg:path" )
                  .attr( "d", "M0,-5L10,0L0,5");
            }

            chart.update = function(source) {
                flow(root);
                var nodes = flow.nodes(),
                    links = flow.links();

                function color(d) {
                  // FIXME remove check for null 
                  // If at least one visible child: white
                  if ( d.children && d.children.length > 0 ) return "white";
                  // If there are any children: lightgray
                  if ( d.allchildren.length > 0 ) return "lightgray";
                  // Otherwise: white
                  return "white";
                }

                function
                toggle( node )
                {
                  // On collapse:
                  if ( node.expanded == true) {
                    node.children = node._children;
                    node._children = null;
                    node.expanded = false;
                  }
                  // On expand
                  else {
                    node._children = node.children;
                    node.children = node.allchildren;
                    node.expanded = true;
                  }
                }

                function expandNode( node ) {
                  switch( node.exp ) {
                    case ExpEnum.ONE:
                      {
                        // If we are expanding this node, collapse all siblings
                        if ( !node.expanded ) {
                          var parent = node.parent;
                          if ( parent ) {
                            // FIXME siblings maybe in vchildren
                            // Get visible siblings 
                            var siblings = parent.children;
                            assert( siblings, "parent should have visible children" ); 

                            // collapse other siblings
                            siblings.forEach( function( sibling ) {
                              if ( sibling.expanded ) toggle( sibling );
                            } );
                          }
                        }
                        toggle( node );
                      }
                      break;
                    
                    case ExpEnum.PASS:
                      {
                        var parent = node.parent;
                        assert( parent != null );
                        expandNode( parent );
                      }
                      break;

                    case ExpEnum.ALL:
                      toggle( node );
                      break;
                    
                    case ExpEnum.OFF:
                    default:
                      {
                        // Do nothing
                      }
                      break;

                  }
                  // FIXME should be resizing the node or its parent?
                  resize( node );
//                  chart.update( node );
                }

                function 
                updateNodes()
                {
                  var node = chartGroup.selectAll("g.node")
                    .data(nodes, function(d) { return d.id2; });

                  var nodeEnter = node.enter().append("svg:g")
                    .attr("class", function( d ) {
                        return "node " + d.shape;
                     } )
                    .attr("transform", function(d) {
                      return "translate(" + source.x + "," + source.y + ")";
                     } )
                    .style("opacity", 1e-6);

                  var rectEnter = nodeEnter.filter( ".RECT" );    
                  var circEnter = nodeEnter.filter( ".CIRC" );

                  // Enter any new nodes at the parent's previous position.
                  rectEnter.append("svg:rect")
                    .attr("class", "background")
                    .attr("height", function(d) { return d.height; })
                    .attr("width", function(d) { return d.width; })
                    .style("fill", color)
                    .on("click", syncPanesToNode)
                    .on("dblclick", expandNode);

                  circEnter.append("svg:circle")
                    .attr("r", function(d) { return d.width/2; })
                    .style("fill", color)
                    .on("click", syncPanesToNode)
                    .on("dblclick", expandNode);

                  nodeEnter.each(function(d) {
                      d3.select(this).append("svg:text")
                      .attr("class", "label")
                      .attr("dy", 10)
                      .attr("dx", 10)
                      .text(function(d) { 
                        if ( d.type == "port" ) return "";
                        return d.name; 
                        });
                      });

                  // This includes the Enter+Update
                  var nodeUpdate = node.transition()
                    .duration(transitionDuration)
                    .attr("transform", function(d) { return "translate(" + d.x + "," + d.y + ")"; });

                  var rectUpdate = nodeUpdate.filter( ".RECT" );
                  var circUpdate = nodeUpdate.filter( ".CIRC" );

                  rectUpdate
                    .style("opacity", 1)
                    .select("rect")
                    .style("fill", color);

                  circUpdate
                    .style("opacity", 1)
                    .select("circle")
                    .style("fill", color);

                  // Assume only rectangular nodes can be background
                  nodeUpdate.select(".background")
                    .attr("height", function(d) { return d.height; })
                    .attr("width", function(d) { return d.width; });

                  // Bring port nodes to front
                  // Assume all circle nodes are ports
                  circUpdate.each( function() {
                    this.parentNode.appendChild( this );
                  } );

                  // Transition exiting nodes to the parent's new position.
                  node.exit().transition()
                    .duration(transitionDuration)
                    .attr("transform", function(d) { return "translate(" + source.x + "," + source.y + ")"; })
                    .style("opacity", 1e-6)
                    .remove();
                }
                
                function
                updateLinks()
                {
////                  console.log( "updateLinks: links: " );
////                  links.forEach( function( link ) {
////                    console.log( link.source.name + " -> " + link.target.name );
////                  } );

                  ///// Line Edges
                  var line = d3.svg.line()
                    .x( function( d ) { return d.x; } )
                    .y( function( d ) { return d.y; } );

                  // NOTE: When this function is called, the axes are already
                  // translate to source's (x,y)
                  function
                  drawLine( link )
                  {
                    var source = {}, target = {};

                    switch( link.source.shape ) {
                      case "CIRC":
                        {
                          source.x = link.source.x - link.source.width/2;
                          source.y = link.source.y;
                        }
                        break;

                      case "RECT":
                        {
                          source.x = link.source.width/2;
                          source.y = link.source.height;
                        }
                        break;

                      default:
                        {
                          source.x = link.source.x;
                          source.y = link.source.y;
                        }
                    }

                    switch( link.target.shape ) {
                      case "CIRC":
                        {
                          target.x = link.target.x - link.target.width/2;
                          target.y = link.target.y;
                        }
                        break;

                      case "RECT":
                        {
                          target.x = link.target.x + link.target.width/2 - link.source.x;
                          target.y = link.target.y - link.source.y;
                        }
                        break;

                      default:
                        {
                          target.x = link.target.x;
                          target.y = link.target.y;
                        }
                    }

                    var points = [ source, target ]; 
                    return line( points ); 
                  }

                  ///// Backedge
                  // TODO cleanup this function
                  // NOTE: When this function is called, the axes are already
                  // translate to the target's (x,y)
                  function
                  getPoints2( link )
                  {
                    var source = {},
                        target = {};

                    assert( link.source.y >= link.target.y );
                    switch( link.source.shape ) {
                      case "RECT":
                        {
                          source.x = link.source.x - link.target.x;
                          source.y = link.source.y + link.source.height - link.target.y;

                          target.x = 0;
                          target.y = 0;

                          return { source:source, target:target };
                        }
                       default:
                        return link;
                    }
                  }

                  function
                  drawBackedge( link )
                  {
                    newLink = getPoints2( link );
                    var dx = newLink.target.x -newLink.source.x;
                    var dy = newLink.target.y - newLink.source.y;
                    var dr = Math.sqrt( dx*dx + dy*dy );
                    return "M" + newLink.source.x + "," + newLink.source.y + "A" + dr + 
                           "," + dr + " 0 0,1 " + newLink.target.x + "," + newLink.target.y;
                  }

                  function
                  getPoint( node, isSource )
                  {
                    switch( node.shape ) {
                      case "CIRC":
                        return { x:node.x - node.width/2, y:node.y }; 

                      case "RECT":
                        {
                          if ( isSource ) return { x:node.x + node.width/2, y:node.y + node.height };
                          else            return { x:node.x + node.width/2, y:node.y };
                        }

                      default:
                        return { x:node.x, y:node.y };
                    }
                  } 

                  var diagonal = d3.svg.diagonal()
                    .source( function( d ) { return getPoint( d.source, true ); } )
                    .target( function( d ) { return getPoint( d.target, false ); } )
                    .projection( function( d ) { return [ d.x, d.y ]; } );
                  
                  function
                  printLink( link )
                  {
////                    console.log( link.source.name + " -> " + link.target.name );
                  }

                  // Update the links
                  var linkUpdate = svg.selectAll( "g.link" )
                    .data( links );

//                  console.log( "linkEnter" );
                  var linkEnter = linkUpdate.enter().append( "svg:g" )
                    .attr( "class", function( d ) {
//                      console.log( "Entering link of type " + d.type );
                      return "link " + d.type;
                    } )
                    .attr( "transform", function( d ) {
                      // TODO turn this into a switch statement
                      // TODO this needs to be refactored because it's called
                      // again below
                      // If this is a forward edge, translate to the source's
                      // position
                      if (    d.type == "CFGFWD" 
                           || d.type == "CONTENTION" ) {
                        return "translate(" + d.source.x + "," + d.source.y + ")";
                      }
                      // If this is a backedge, translate to the target's
                      // position
                      if ( d.type == "CFGBWD" ) {
                        return "translate(" + d.target.x + "," + d.target.y + ")";
                      }
                      if ( d.type == "MEMDEP" ) {
                        var parent = d.source.parent;
                        return "translate(" + parent.x + "," + d.source.y + ")";
                      }
                      return "translate(0,0)";
                    } )
                    .style( "opacity", function( d ) {
                      if ( d.type === "MEMDEP" )
                        return 0.6;
                      return 1.0;
                    } );
                  
                  var rectEnter = linkEnter.filter( function( d ) {
                    if ( d.type === "MEMDEP" ) return true;
                    return false;
                  } );
                  rectEnter.append( "svg:rect" )
                    .attr( "width", function( d ) { 
                      var parent = d.source.parent;
                       return parent.width;
                    } )
                    .attr( "height", function( d ) {
                      return d.target.y - d.source.y + d.target.height;
                    } )
                    .style( "fill", "gold" );
                  
                  var contEnter = linkEnter.filter( function( d ) {
                    if ( d.type === "CONTENTION" ) return true;
                    return false;
                  } );
                  contEnter
                    .append( "path" )
                    .attr( "d", function( d ) {
                      var node = { x:source.x0, y:source.y0 };
                      return drawLine( { source:node, target:node } );
                    } )
                    .attr( "marker-end", "url(#end)" );
                  
                  var cfgbwdEnter = linkEnter.filter( function( d ) {
                    if ( d.type === "CFGBWD" ) return true;
                    return false;
                  } );
 
                  var cfgfwdEnter = linkEnter.filter( function( d ) {
                    if ( d.type === "CFGFWD" ) return true;
                    return false;
                  } );

                  // Enter any new links at the parent's pervious position.
                  cfgbwdEnter
                    .append( "path" )
                    .attr( "d", function( d ) {
                      var node  = { x:source.x0, y:source.y0 };
                      return drawBackedge( { source:node, target:node } );
                    } )
                    .attr( "marker-end", "url(#end)" );
                  
                  cfgfwdEnter
                    .append( "path" )
                    .attr( "d", function( d ) {
                      var node = { x:source.x0, y:source.y0 };
                      return drawLine( { source:node, target:node } );
                    } )
                    .attr( "marker-end", "url(#end)" );

                  linkUpdate2 = linkUpdate
                    .transition()
                    .duration( transitionDuration );
                  linkUpdate2  
                    .attr( "class", function( d ) {
//                      console.log( "Updating link of type " + d.type );
                      return "link " + d.type;
                    } )
                    .attr( "transform", function( d ) {
                      // TODO turn this into a switch statement, possibly make
                      // it a named function
                      // If this is a forward edge, translate to the source's
                      // position
                      if (    d.type == "CFGFWD" 
                           || d.type == "CONTENTION" ) {
                        return "translate(" + d.source.x + "," + d.source.y + ")";
                      }
                      // If this is a backedge, translate to the target's
                      // position
                      if ( d.type == "CFGBWD" ) {
                        return "translate(" + d.target.x + "," + d.target.y + ")";
                      }
                      if ( d.type == "MEMDEP" ) {
                        var parent = d.source.parent;
                        return "translate(" + parent.x + "," + d.source.y + ")";
                      }
                      return "translate(0,0)";
                    } ) 
                    .style( "opacity", function( d ) {
                      if ( d.type === "MEMDEP" )
                        return 0.6;
                      return 1.0;
                    } ); 
                  
                  var rectUpdate = linkUpdate2.filter( function( d ) {
                    if ( d.type === "MEMDEP" ) return true;
                    return false;
                  } )
                  .selectAll( "rect" );
//                    .attr("transform", function(d) { return "translate(" + d.x + "," + d.y + ")"; });
                  //TODO
                
                  var contUpdate = linkUpdate2.filter( function( d ) {
                    if ( d.type === "CONTENTION" ) return true;
                    return false;
                  } );
//                  console.log( "contUpdate:" );
//                  console.log( contUpdate );  
                  contUpdate
                    .selectAll( "path" )
                    .attr( "d", drawLine );
                  
                  var cfgbwdUpdate = linkUpdate2.filter( function( d ) {
                    if ( d.type === "CFGBWD" ) return true;
                    return false;
                  } ).selectAll( "path" );
                  var cfgfwdUpdate = linkUpdate2.filter( function( d ) {
                    if ( d.type === "CFGFWD" ) return true;
                    return false;
                  }).selectAll( "path" );

                  // Transition links to their new position.
                  cfgbwdUpdate
                    .attr( "d", drawBackedge );

                  cfgfwdUpdate
                    .attr( "d", drawLine);

                  // Transition exiting nodes to the parent's new position.
                  var linkExit = linkUpdate.exit();
                  
                  var rectgExit = linkExit.filter( function( d ) {
                    if ( d.type === "MEMDEP" ) return true;
                    return false;
                    } )
                    .transition()
                    .duration( transitionDuration );
                  var rectExit  = rectgExit.selectAll( "rect" );

                  rectgExit.remove(); 
                  
                  var contgExit = linkExit.filter( function( d ) {
                    if ( d.type === "CONTENTION" ) return true;
                    return false;
                    } )
                    .transition()
                    .duration( transitionDuration );

                  var contExit = contgExit.selectAll( "path" );
                  contExit
                    .attr( "d", function( d ) {
                        var node = { x:source.x, y:source.y };
                        return drawLine( { source:node, target:node } );
                    } );

                  contgExit.remove();  

                  var cfgbwdgExit = linkExit.filter( function( d ) {
                    if ( d.type === "CFGBWD" ) return true;
                    return false;
                    } )
                    .transition()
                    .duration( transitionDuration );

                  var cfgfwdgExit = linkExit.filter( function( d ) {
                    if ( d.type === "CFGFWD" ) return true;
                    return false;
                    } )
                    .transition()
                    .duration( transitionDuration );

                  var cfgbwdExit = cfgbwdgExit.selectAll( "path" );
                  var cfgfwdExit = cfgfwdgExit.selectAll( "path" );
                 
                  cfgbwdExit 
                    .attr( "d", function( d ) {
                        var node = { x:source.x, y:source.y };
                        return drawBackedge( { source:node, target:node } );
                    } );
                  
                  cfgfwdExit
                    .attr( "d", function( d ) {
                      var node = { x:source.x, y:source.y };
                      return drawLine( { source:node, target:node } );
                    } );

                  cfgbwdgExit.remove();  
                  cfgfwdgExit.remove();  
                }

                function
                updateLinks2()
                {
                  // TODO rename
                  // NOTE: all data have to be of the same type
                  function
                  deal( links, type )
                  {
                    var linkType = "NOTSET";  
                    links.forEach( function( d ) {
                      if ( linkType === "NOTSET" )
                        linkType = d.type;
                      assert( linkType === d.type );  
                    } );
                    assert(    linkType === "NOTSET"
                            || linkType === type );
                    
//                    links.forEach( printLink ); 
                    
                    var selection = chartGroup.selectAll( "g.link." + type )
                      .data( links, function( d ) { return ( d.source.id2 * maxID2 + d.target.id2 ); } );
                    
                    // ENTER
                    var enter = selection.enter().append( "g" )
                      .attr( "class", function( d ) {
                        return "link " + d.type;
                      } )
                      .style( "opacity", 1e-6 );
                  
                    switch( type ) {
                      case "CFGFWD":
                      {
                        enter.append( "path" )
                          .attr( "d", function( d ) {
                            var node = { x:source.x0, y:source.y0 };
                            var link = { source:node, target:node };
                            return drawLine( link );
                          } )
                          .attr( "marker-end", "url(#end)" );
                      }
                      break;
                      case "CFGBWD":
                      {
                        enter.append( "path" )
                          .attr( "d", function( d ) {
                            var node = { x:source.x0, y:source.y0 };
                            var link = { source:node, target:node };
                            return drawBackedge( link );
                          } )
                          .attr( "marker-end", "url(#end)" );
                      }
                      break; 
                      case "CONTENTION":
                      {
                        enter.append( "path" )
                          .attr( "d", function( d ) {
                            var node = { x:source.x0, y:source.y0 };
                            var link = { source:node, target:node };
                            return drawLine( link );
                          } )
                          .attr( "marker-end", "url(#end)" );
                      }
                      break;
                      // TODO change this to use diagonal
                      case "PORT":
                      {
                        enter.append( "path" )
                          .attr( "d", function( d ) {
                            var node = { x:source.x0, y:source.y0 };
                            var link = { source:node, target:node };
                            return diagonal( link );
                          } );
                      }
                      break;
                      default:
                    }
                    
                    // UPDATE
                    update = selection.transition()
                      .duration( transitionDuration )
                      .attr( "transform", function( d ) { 
                        // TODO turn this into a switch statement
                        // TODO this needs to be refactored because it's called
                        // again below
                        // If this is a forward edge, translate to the source's
                        // position
                        if (    d.type === "CFGFWD" 
                             || d.type === "CONTENTION" ) {
                          return "translate(" + d.source.x + "," + d.source.y + ")";
                        }
                        // If this is a backedge, translate to the target's
                        // position
                        if ( d.type === "CFGBWD" ) {
                          return "translate(" + d.target.x + "," + d.target.y + ")";
                        }
                        if ( d.type === "MEMDEP" ) {
                          var parent = d.source.parent;
                          return "translate(" + parent.x + "," + d.source.y + ")";
                        }
//                        if ( d.type === "PORT" ) {
//                          return "translate(" + d.source.x + "," + d.source.y + ")";
//                        }
                        return "translate(0,0)"; 
                      } );

                    switch( type ) {
                      case "CFGFWD":
                      {
                        update
                          .style( "opacity", 1 )
                          .selectAll( "path" )
                          .attr( "d", drawLine );
                      }
                      break;
                      case "CFGBWD":
                      {
                        update
                          .style( "opacity", 1 )
                          .selectAll( "path" )
                          .attr( "d", drawBackedge );
                      }
                      break;
                      case "CONTENTION":
                      {
                        update
                          .style( "opacity", 1 )
                          .selectAll( "path" )
                          .attr( "d", drawLine );
                      }
                      break;
                      case "PORT":
                      {
                        update
                          .style( "opacity", 1 )
                          .selectAll( "path" )
                          .attr( "d", diagonal );
                      }
                      break;
                      default:
                    }

                    // EXIT
                    var exit = selection.exit().transition()
                      .duration( transitionDuration )
                      .style( "opacity", 1e-6)
                      ;

                    switch( type ) {
                      case "CFGFWD":
                      {
                        exit.selectAll( "path" ) 
                          .attr( "d", function( d ) {
                            var node = { x:source.x, y:source.y };
                            var link = { source:node, target:node };
                            return drawLine( link );
                          } );
                      }
                      break;
                      case "CFGBWD":
                      {
                        exit.selectAll( "path" ) 
                          .attr( "d", function( d ) {
                            var node = { x:source.x, y:source.y };
                            var link = { source:node, target:node };
                            return drawBackedge( link );
                          } );
                      }
                      break;
                      case "CONTENTION":
                      {
                        exit.selectAll( "path" ) 
                          .attr( "d", function( d ) {
                            var node = { x:source.x, y:source.y };
                            var link = { source:node, target:node };
                            return drawLine( link );
                          } );
                      }
                      break;
                      case "PORT":
                      {
                        exit.selectAll( "path" ) 
                          .attr( "d", function( d ) {
                            var node = { x:source.x, y:source.y };
                            var link = { source:node, target:node };
                            return diagonal( link );
                          } );
                      }
                      break;
                      default:
                    }
                    
                    exit.remove();  
                  }


//                  console.log( "updateLinks2: links: " );
//                  links.forEach( function( link ) {
//                    console.log( link.source.name + " -> " + link.target.name );
//                  } );

                  ///// Line Edges
                  var line = d3.svg.line()
                    .x( function( d ) { return d.x; } )
                    .y( function( d ) { return d.y; } );

                  // NOTE: When this function is called, the axes are already
                  // translate to source's (x,y)
                  function
                  drawLine( link )
                  {
                    var source = {}, target = {};

                    switch( link.source.shape ) {
                      case "CIRC":
                        {
                          source.x = link.source.width/2;
                          source.y = link.source.height;
                        }
                        break;

                      case "RECT":
                        {
                          source.x = link.source.width/2;
                          source.y = link.source.height;
                        }
                        break;

                      default:
                        {
                          source.x = link.source.x;
                          source.y = link.source.y;
                        }
                    }

                    switch( link.target.shape ) {
                      case "CIRC":
                        {
                          target.x = link.target.x + link.target.width/2 - link.source.x;
                          target.y = link.target.y - link.source.y;
                        }
                        break;

                      case "RECT":
                        {
                          target.x = link.target.x + link.target.width/2 - link.source.x;
                          target.y = link.target.y - link.source.y;
                        }
                        break;

                      default:
                        {
                          target.x = link.target.x;
                          target.y = link.target.y;
                        }
                    }

                    var points = [ source, target ]; 
                    return line( points ); 
                  }

                  ///// Backedge
                  // TODO cleanup this function
                  // NOTE: When this function is called, the axes are already
                  // translate to the target's (x,y)
                  function
                  getPoints2( link )
                  {
                    var source = {},
                        target = {};

//                    console.log( "getPoints2 on: " );
//                    console.log( link.source.name + " -> " + link.target.name );
                    assert( link.source.y >= link.target.y );
                    switch( link.source.shape ) {
                      case "RECT":
                        {
                          source.x = link.source.x - link.target.x;
                          source.y = link.source.y + link.source.height - link.target.y;

                          target.x = 0;
                          target.y = 0;

                          return { source:source, target:target };
                        }
                       default:
                        return link;
                    }
                  }

                  function
                  drawBackedge( link )
                  {
                    newLink = getPoints2( link );
                    var dx = newLink.target.x -newLink.source.x;
                    var dy = newLink.target.y - newLink.source.y;
                    var dr = Math.sqrt( dx*dx + dy*dy );
                    return "M" + newLink.source.x + "," + newLink.source.y + "A" + dr + 
                           "," + dr + " 0 0,1 " + newLink.target.x + "," + newLink.target.y;
                  }

                  function
                  getPoint( node, isSource )
                  {
                    switch( node.shape ) {
                      case "CIRC":
                        return { x:node.x - node.width/2, y:node.y }; 

                      case "RECT":
                        {
                          if ( isSource ) return { x:node.x + node.width/2, y:node.y + node.height };
                          else            return { x:node.x + node.width, y:node.y + node.height/2 };
                        }

                      default:
                        return { x:node.x, y:node.y };
                    }
                  } 

                  var diagonal = d3.svg.diagonal()
                    .source( function( d ) { return getPoint( d.source, true ); } )
                    .target( function( d ) { return getPoint( d.target, false ); } )
                    .projection( function( d ) { return [ d.x, d.y ]; } );
                  
                  function
                  printLink( link )
                  {
////                    console.log( link.source.name + " -> " + link.target.name );
                  }

                  // NEW!
                  var CFGFWDLinks     = links.filter( function( d ) { return ( d.type === "CFGFWD" ) } );
                  var CFGBWDLinks     = links.filter( function( d ) { return ( d.type === "CFGBWD" ) } );
                  var MEMDEPLinks     = links.filter( function( d ) { return ( d.type === "MEMDEP" ) } );
                  var CONTENTIONLinks = links.filter( function( d ) { return ( d.type === "CONTENTION" ) } );
                  var PORTLinks       = links.filter( function( d ) { return ( d.type === "PORT") } );

//                  console.log( "------------------------------");
//                  console.log( "CFGFWDLinks: " ); 
//                  CFGFWDLinks.forEach( printLink );
//                  console.log( "------------------------------");
//
//                  console.log( "------------------------------");
//                  console.log( "CFGBWDLinks: " ); 
//                  CFGBWDLinks.forEach( printLink );
//                  console.log( "------------------------------");
//
//                  console.log( "------------------------------");
//                  console.log( "MEMDEPLinks: " ); 
//                  MEMDEPLinks.forEach( printLink );
//                  console.log( "------------------------------");
//
//                  console.log( "------------------------------");
//                  console.log( "CONTENTIONLinks: " ); 
//                  CONTENTIONLinks.forEach( printLink );
//                  console.log( "------------------------------");

                  deal( CFGFWDLinks,      "CFGFWD" );  
                  deal( CFGBWDLinks,      "CFGBWD" );  
                  deal( CONTENTIONLinks,  "CONTENTION" );
                  deal( [],        "PORT" );
                  deal( PORTLinks,        "PORT" );
                }

                updateNodes();
                updateLinks2();

                // Remember old positions for transition
                nodes.forEach( function( d ) {
                  d.x0 = d.x;
                  d.y0 = d.y;
                } );
            };

            root.x0 = 0, root.y0 = 0;
            resize(root);
            chart.update(root);
        });
    }

    chart.width = function(value) {
        if (!arguments.length) return width;
        width = parseInt(value);
        return this;
    };

    chart.height = function(value) {
        if (!arguments.length) return height;
        height = parseInt(value);
        return this;
    };

    chart.margin = function(_) {
        if (!arguments.length) return margin;
        margin.top    = typeof _.top    != 'undefined' ? _.top    : margin.top;
        margin.right  = typeof _.right  != 'undefined' ? _.right  : margin.right;
        margin.bottom = typeof _.bottom != 'undefined' ? _.bottom : margin.bottom;
        margin.left   = typeof _.left   != 'undefined' ? _.left   : margin.left;
        return chart;
    };

    return chart;
};
