d3.custom.layout.flow = 
function() 
{
  var hierarchy = d3.layout.hierarchy().sort( null ).value( null ),
      width     = 0,
      height    = 0,
      padding   = { top:20, left:10, bottom:10, right:10 },
      margin    = { top:10, left:10, bottom:10, right:10 };

  var nodes = [],
      allnodes = [],
      links = [],
      visibleNodeIds = {},
      oldVisibleNodeIds = {};
  
  // TODO Unused variables
  var nodeWidth = 0,
      nodeHeight = 0,
      containerHeight = 0;
  
  // Replace all uses of nodeWidth and nodeHeight with node.iw and node.ih
  // containerWidth and containerHeight have to be calculated based on whether
  // the nodes is showing its children or not

  function 
  flow( d, i )
  {
    ///// Printing functions
    function
    getNodeName( node )
    {
      return ( node.name + "(" + node.id2 + ")" );
    }
    
    function
    printNode( node )
    {
      console.log( getNodeName( node ) );
    }

    function
    printLink( link )
    {
      var sourceName = getNodeName( link.source );
      var targetName = getNodeName( link.target );
      console.log( sourceName   + " -> " + targetName );
    }

    function
    printEdge( edge )
    {
      var fromName = getNodeName( nodeIdMap[ edge.from ] );
      var toName = getNodeName( nodeIdMap[ edge.to ] );
      console.log( fromName   + " -> " + toName );
    } 
    
    // Prints map< obj, array >
    function
    printMap( map )
    {
      for( var key in map ) {
        if ( map.hasOwnProperty( key ) ) {
          var obj = nodeIdMap[ key ]; 

          var value = map[ key ];
          assert( value instanceof Array );
          console.log( getNodeName( obj ) + " -> [" );
          value.forEach( printEdge );
          console.log( "]" );
        }
      }
    }

    // Prints map< array, obj >
    function
    printFlippedMap( map )
    {
      for( var key in map ) {
        if ( map.hasOwnProperty( key ) ) {
          var obj = nodeIdMap[ map[ key ] ]; 
          console.log( "[" + key + "] -> " + getNodeName( obj ) );
        }
      }
     }
    
    ///// Other functions
    function
    gatherAllNodes( node )
    {
      allnodes.push( node );
      node.allchildren.forEach( gatherAllNodes ); 
    }

    function
    gatherVisibleNodes( node )
    {
      nodes.push( node );
      var children = getChildren( node );
      children.forEach( gatherVisibleNodes );
    }

    function
    gatherVisibleNodeIds( nodes )
    {
      // Clear out the map
      visibleNodeIds = {};

      nodes.forEach( function( node ) {
        visibleNodeIds[ node.id2 ] = true;
      } );
    }

    function
    findLinksForNode( node )
    {
      // Put the node in an array so that we can use "map"
      var nodes = [];
      nodes.push( node );
      // gather all links for visible nodes
      var newLinks = d3.merge( nodes.map( function( node ) {
        return ( node.edges || [] ).map( function( edge ) {
          // TODO should we pass nodeIdMap to the constructor?
          return { source: nodeIdMap[ edge.from ], target: nodeIdMap[ edge.to ] };
        } )
      } ) );

      return newLinks;
    }

    // TODO rename
    function
    sortLinkOnid2( link0, link1 )
    {
      return ( link0.target.id2 - link1.target.id2 );
    }

    function
    sortOnSourceThenTarget( link0, link1 )
    {
      if ( link0.source.id2 != link1.source.id2 ) {
        return ( link0.source.id2 - link1.source.id2 );
      }
      return ( link0.target.id2 - link1.target.id2 );
    }


    function
    gatherVisibleLinks( nodes )
    {
//      console.log( "gathering visible links" );
      // gather all links for visible nodes
      var allLinks = d3.merge( nodes.map( function( node ) {
        var edges = [];
        // Take virtual edges over normal edges
        if ( node.vedges ) edges = edges.concat( node.vedges );
        if ( node.edges )  edges = edges.concat( node.edges );
        
        return ( edges ).map( function( edge ) {
          // TODO should we pass nodeIdMap to the constructor?
          return { source: nodeIdMap[ edge.from ], target: nodeIdMap[ edge.to ], type: edge.type };
        } )
      } ) );

      // Go through all the links and only keep the ones where the target is
      // also visible 
      var visibleLinks = [];
      allLinks.forEach( function( link ) {
        // Otherwise, show the link only if the target is also visible
        if ( visibleNodeIds[ link.target.id2 ] == true )
          visibleLinks.push( link );
      } );

      return visibleLinks;
    }

    function
    getChildren( node )
    {
      if ( node.vchildren ) {
        // if this node had visible children, then the vchildren would have
        // been added to that
        assert( !node.children );
        return node.vchildren;
      }
      var children = [];
      if ( node.children ) children = node.children;
      return children;
    }
    
    function
    firstWalk( node )
    {
      var children = getChildren( node );
      // The node has visible children
      if ( children.length > 0 ) {
        children.forEach( firstWalk );
        layoutNode( node );
      }
      else {
        node.x = 0;
        node.y = 0;
        node.width = node.iw;
        node.height = node.ih;
      }
    }

    function
    secondWalk( node )
    {
      var children = getChildren( node );
      if ( children.length == 0 ) return;
      children.forEach( function( child ) {
        child.x += node.x;
        child.y += node.y;
        secondWalk( child );
      } );
    }

    // Layout a node and its children
    // This function should only be called on nodes that have (visible)
    // children
    function
    layoutNode( node )
    {
      // Lay things out differently based on the child type
      var children = getChildren( node );
      if ( children.length > 0 ) {
        // Assume homogeneous children
        var childType = "NOTSET";
        children.forEach( function( child )  {
          if ( childType === "NOTSET" ) childType = child.type;
          assert( child.type === childType );
        } );
        switch( childType ) {
          case "bank":
            layoutMemSystem( node );
            break;
          case "port":
            layoutBank( node );
            break;
          default:
            gridLayout( node );
            break;
        }
      }
    }

    function
    layoutMemSystem( node )
    {
      var   children  = getChildren( node ) 
          , curX      = 0     
          , curY      = 0     
          , width     = 0     
          ;

      children.forEach( function( child ) {
        child.x = curX;
        child.y = curY;
      
        curY += child.height;

        // Ensure all children have the same width
        if ( width == 0 ) width = child.width;
        else              assert( width == child.width );
      } );    

      node.width = width;
      node.height = curY;
      
      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }

    function
    layoutBank( node )
    {
      var   children  = getChildren( node ) 
          , curY      = padding.top/2 // TODO 
          ;
      
      children.forEach( function( child ) {
        child.x = 0;
        child.y = curY + child.height/2;

        curY += child.height/2 + padding.bottom;

        // Ensure that the child has a smaller width than this node
        assert( child.width < node.iw );
      } );

      node.width = node.iw;
      node.height = Math.max( curY, node.ih ); 

      node.x = 0;
      node.y = 0;
    }

    function
    gridLayout( node )
    {
      var   children  = getChildren( node ) 
          , curX      = 0 
          , curY      = 0 
          , maxWidth  = 0
          , maxHeight = 0
          , prev      = null;
      
      children.forEach( function( child ) {
        if ( prev ) {
          if ( prev.order == child.order ) {
            // same row, next column
            curX += prev.width + padding.right;
          }
          else {
            // next row, column 0
            curY += prev.height + padding.bottom;
            curX = 0;
          }
        }

        // (curX, curY) is the starting point of the cell
        // Move to the middle of the cell
        child.x = curX + padding.left;
        child.y = curY + padding.top;
        
        maxWidth = Math.max( maxWidth, child.x + child.width + padding.right );
        maxHeight = Math.max( maxHeight, child.y + child.height + padding.bottom );
        prev = child;
      } );

      node.width = maxWidth;
      node.height = maxHeight;

      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }

    ///// NEW FUNCTIONS
    
    // Sorting functions
    function
    sortEdgeOnTarget( edge0, edge1 )
    {
      return ( edge0.to - edge1.to );
    }

    function
    sortOnId2( node0, node1 )
    {
      return ( node0.id2 - node1.id2 );
    }
    
    // Utility functions
    // NOTE: this assumed the edge sources are all the same and doesn't check
    // for them!
    function
    removeDuplicateEdges( sortedEdges )
    {
      var uniqued = [];
      var prev = -1;
      sortedEdges.forEach( function( edge ) {
        if ( edge.to != prev ) {
          uniqued.push( edge );
          prev = edge.to;
        }
      } ); 
      return uniqued;
    }
    
    // Handles array and objects
    function
    deepCopy( obj )
    {
      if ( obj instanceof Array ) {
        var copy = [];
        for ( var i = 0, e = obj.length; i != e; ++i ) {
          copy[ i ] = arguments.callee( obj[ i ] );
        }
        return copy;
      }
      
      if ( typeof obj === "object" ) {
        var copy = {};
        for ( var i in obj ) {
          copy[ i ] = arguments.callee( obj[ i ] );
        }
        return copy;
      }
      
      return obj;
    }

    // Functions related to virtual links
    function
    resetVirtuals( allnodes )
    {
      allnodes.forEach( function( node ) {
        if ( node.vchildren ) node.vchildren = null;
        if ( node.vedges ) node.vedges = null;
        if ( node.vparent ) node.vparent = null;
      } );
    }


    function
    gatherHiddenPortEdges( node, portToEdgesMap )
    {
      var edges = ( node.edges ? node.edges : [] );
      var portEdges = edges.filter( function( d ) {
        // If this is a port edge where one endpoint is hidden
        var sourceVisible = ( visibleNodeIds[ d.from ] === true );
        var targetVisible = ( visibleNodeIds[ d.to ] === true );
        return (    d.type === "PORT"
                 && ( !sourceVisible || !targetVisible ) );
      } );
      if ( portEdges.length > 0 ) {
        // Make a deep copy of the edges array, so that when we update the edge
        // endpoints it doesn't touch the original edge objects
        var copy = deepCopy( portEdges );
        portToEdgesMap[ node.id2 ] = copy;
      }
      node.allchildren.forEach( function( child ) {
        gatherHiddenPortEdges( child, portToEdgesMap );
      } );
    }
    
    function
    updatePortEdgeEndPoints( portToEdgesMap )
    {
      for( var portId in portToEdgesMap ) {
        if( !portToEdgesMap.hasOwnProperty( portId ) ) continue;
        
        var portEdges = portToEdgesMap[ portId ];
        assert( portEdges instanceof Array );
        portEdges.forEach( function( edge ) {
          assert( portId == edge.from );

          // TODO it may make more sense to operate on links, so we don't have
          // to do the nodeIdMap lookup everytime
          // Make the edge target the first visible parent
          while( visibleNodeIds[ edge.to ] != true ) {
            var target = nodeIdMap[ edge.to ];
            edge.to = target.parent.id2;
          }
        } );

        // Now we have a set of edges from port -> visible nodes
        // If there are duplicate edges, remove them here
        portEdges.sort( sortEdgeOnTarget );
        portToEdgesMap[ portId ] = removeDuplicateEdges( portEdges );  
      }
    }

    // Accepts an array of unsigned integers
    function
    verifyIsSorted( array )
    {
      var prev = -1;
      array.forEach( function( n ) {
        // Note: we assume the array doesn't contain duplicate elements
        assert( n > prev );
        prev = n;
      } );
    }

    // Note: The edges in the map need to be sorted
    function
    flipPortToEdgesMap( map )
    {
      var flipped = {};
      for ( var key in map ) {
        if ( !map.hasOwnProperty( key ) ) continue;

        var value = map[ key ];
        assert( value instanceof Array );

        var targetIds = value.map( function( edge ) {
          return edge.to;
        } );
//        console.log( "targetids: " );
//        console.log( targetIds );
        
        verifyIsSorted( targetIds );
        
        // This works because target ids are sorted
        flipped[ targetIds ] = key;
      }
      
      return flipped;
    }
    
    function
    getVirtualPortEdges( portToEdgesMap )
    {
      var first = true;
      var portsAreVisible;
      for ( var portId in portToEdgesMap ) {
        if ( !portToEdgesMap.hasOwnProperty( portId ) ) continue; 
        
        if ( first ) portsAreVisible = ( visibleNodeIds[ portId ] == true );
        assert( portsAreVisible == ( visibleNodeIds[ portId ] == true ) );

        first = false;
      }
      
      var virtualEdges = [];
      // If the ports are visible, don't cluster/filter edges, show them all
      if ( portsAreVisible ) {
       for ( var portId in portToEdgesMap ) {
         if ( !portToEdgesMap.hasOwnProperty( portId ) ) continue; 
         portEdges = portToEdgesMap[ portId ];
         assert( portEdges instanceof Array );
         virtualEdges = virtualEdges.concat( portEdges );
       }      
      }
      else {
        var nodeSetToPortMap = flipPortToEdgesMap( portToEdgesMap );
        for ( var key in nodeSetToPortMap ) {
          if ( !nodeSetToPortMap.hasOwnProperty( key ) ) continue;

          var portId = nodeSetToPortMap[ key ];
          var portEdges = portToEdgesMap[ portId ];
          assert( portEdges instanceof Array );
          virtualEdges = virtualEdges.concat( portEdges );
        }
      }

      return virtualEdges;
    }
    
    function
    addToVirtualEdges( edges )
    {
      edges.forEach( function( edge ) {
        var source = nodeIdMap[ edge.from ];
        if ( !source.vedges ) source.vedges = [];
        source.vedges.push( edge );
//        console.log( getNodeName( source ) + " # vedges: " + source.vedges.length );
      } );
    }

    function
    addToVirtualChildren( vedges )
    {
      vedges.forEach( function( vedge ) {
        // If the source is visible, nothing to do
        if ( visibleNodeIds[ vedge.from ] == true ) return;
        var source = nodeIdMap[ vedge.from ];

        // We've already added this node to a visible parent
        if ( source.vparent != null ) return;
        
        var visibleParent = source.parent;
        while( visibleNodeIds[ visibleParent.id2 ] != true )
          visibleParent = visibleParent.parent;
        
        if ( !visibleParent.vchildren ) visibleParent.vchildren = [];
        var vchildren = visibleParent.vchildren;
//        console.log( visibleParent.vchildren );
//        console.log( "Adding " + getNodeName( source ) + 
//                     " to vchildren of " + getNodeName( visibleParent ) );
        vchildren.push( source );
        source.vparent = visibleParent;
      } );
    }

    // Arguments:
    // node: Start with the root node and call recursively
    // Returns:
    // Array of visible source nodes
    function
    gatherVisiblePortLinkSources( node )
    {
      var children = node.children;
      // If the node has no visible children or it is not expanded, stop here
      if ( !children || !( node.expanded ) ) {
        visiblePortLinkSources.push( node ); 
        return;
      }
      children.forEach( gatherVisiblePortLinkSources );
    }


    // If the node is visible, return an empty array 
    // If it's hidden it will return the array of nodes whose children need 
    // to be come visible for this node to be shown
    function
    findNode( node )
    {
//      console.log( "Finding node: " );
//      printNode( node );
      
      var allParents = [];
      if ( visibleNodeIds[ node.id2 ] == true ) return allParents;

//      console.log( "node wasn't visible." );
      var nodeParent = node.parent;
      while ( nodeParent != null ) {
        allParents.push( nodeParent );
//        console.log( "Adding parent " + nodeParent.name );
        
        // Arrived at a visible node 
        if ( visibleNodeIds[ nodeParent.id2 ] == true ) break;
        nodeParent = nodeParent.parent;
      }
      
      return allParents;
    }
    
    // Returns an array of new nodes that need to be become visible
    // because they are/container the target of this node's edges
    function
    findNewNodesForLinks( node )
    {
//      console.log( "Finding new links for node " + node.name )
      // Put the node in an array so that we can use "map"
      var nodes = [];
      var newNodes = [];
      nodes.push( node );
      // gather all links for visible nodes
      var newLinks = d3.merge( nodes.map( function( node ) {
        return ( node.edges || [] ).map( function( edge ) {
          // FIXME should we pass nodeIdMap to the constructor?
          return { source: nodeIdMap[ edge.from ], target: nodeIdMap[ edge.to ] };
        } )
      } ) );

      // For all new links, figure out which new nodes need to become
      // visible
      newLinks.forEach( function( link ) 
      {
//        console.log( "Turn on nodes for link:" );
//        printLink( link );
        var toTurnOn = findNode( link.target );
        toTurnOn.forEach( function( d ) {
          newNodes.push( d );
          d._children = d.children;
          d.children = d.allchildren;
          d.expanded = true;
        } );
      } );

      return newNodes;
    }

    // START HERE 

    // Reset stuff
    oldVisibleNodeIds = visibleNodeIds;
    
    nodes = hierarchy.call( this, d, i );
    gatherVisibleNodeIds( nodes );
    var root  = nodes[ 0 ];
    links = [];

    var enableAutomaticExpansion = false;
    if ( enableAutomaticExpansion ) {
      console.log( "*********************************" );
    
     // Figure out which of the nodes are newly added
     var newNodes = [];
     nodes.forEach( function( node )
     {
       if ( oldVisibleNodeIds[ node.id2 ] != true )
         newNodes.push( node );
     } );

     // For all newly added nodes, collect the edges
     // More nodes may have to be added based on what edges we want
     // to show, so nodes.length grows within the for loop
     for ( var i = 0; i < newNodes.length; ++i ) {
       var newNode = newNodes[ i ];
       var newNodesForLinks = findNewNodesForLinks( newNode );
       newNodesForLinks.forEach( function( d ) {
         newNodes.push( d );
       } );
     }
 
    }

    // TODO this should be done only once
    allnodes = [];
    gatherAllNodes( root );

//    console.log( "All Nodes: " );
//    allnodes.forEach( printNode );

//    var enableClustering = true;
    if ( !disableClusterEdgesFlag )
    {
      function
      initVirtualEdges( root )
      {
//        console.log( "+++++++++++++++++++++++++++++++++++++" );
//        console.log( "Starting with root as: " );
//        printNode( root );

        // 1. gather hidden port edges 
        var portToEdgesMap = {};
        gatherHiddenPortEdges( root, portToEdgesMap );
//        console.log( "Port->Edges Map:(at least 1 endpoint is hidden)" );
//        printMap( portToEdgesMap );
        
        // 2. Find the virtual links
        // 2a. Update the endpoints for the edges
        //     Also removes any duplicate edges
        updatePortEdgeEndPoints( portToEdgesMap );
//        console.log( "Updated endpoints/uniqued Port->Edges Map: " );
//        printMap( portToEdgesMap );
    
        // 2b. cluster edges, only if port is not visible 
        var virtualPortEdges = getVirtualPortEdges( portToEdgesMap );
//        console.log( "Virtual Edges:" );
//        virtualPortEdges.forEach( printEdge );


        // 3a. Add the virtual edges to source.vedges
        addToVirtualEdges( virtualPortEdges );
        // 3b. For each virtual edge port's source:
        //     Find its visible parent, and add it as a vchild
        //     We also set the visible parent as the child's vparent (this is so
        //     that we don't add the node as a child twice)
        addToVirtualChildren( virtualPortEdges );

//        console.log( "Done with root as: " );
//        printNode( root );
//        console.log( "+++++++++++++++++++++++++++++++++++++" );
      }
      
//      console.log( "================================" );
      // In the Memory System Graph:
      // Traverse from top down and stop at expanded nodes
      // This will give us a memory type, memory system, bank, or port node
      // These are the visible endpoints of port links
      var memSysGraph = root.children[ 1 ];
      assert( memSysGraph != null );
      
      resetVirtuals( allnodes );
      
      var visiblePortLinkSources = [];
      gatherVisiblePortLinkSources( memSysGraph );
//      console.log( "Visible Port Link Sources:" );
//      visiblePortLinkSources.forEach( printNode );

      visiblePortLinkSources.forEach( function( system ) {
          initVirtualEdges( system ); 
      } );

//      console.log( "DONE" );
//      console.log( "================================" );
    }


    // Now, gather all nodes. We look at both children and vchildren
    // Also sort them so that the children are not hidden behind their parents
    // in the graph
    nodes = [];
    gatherVisibleNodes( d );

//    console.log( "All flow nodes: [" );
//    nodes.forEach( printNode );
//    console.log( "]" );

    // Now we know all visible nodes, remember their ids 
    gatherVisibleNodeIds( nodes );

    // Add all links for visible nodes. If a link's target is 
    // still not visible, then we don't add it
    links = gatherVisibleLinks( nodes );

//    console.log( "Final links: [" );
//    links.forEach( printLink );
//    console.log( "]" );

    firstWalk( root );
    secondWalk( root );
    width = root.width;
    height = root.height; 

    return nodes;
  }

  flow.padding =
  function( _ )
  {
    if ( !arguments.length ) return padding;
    padding.top     = typeof _.top    != 'undefined'  ? _.top     : padding.top;
    padding.right   = typeof _.right  != 'undefined'  ? _.right   : padding.right;
    padding.bottom  = typeof _.bottom != 'undefined'  ? _.bottom  : padding.bottom;
    padding.left    = typeof _.left   != 'undefined'  ? _.left    : padding.left;
    return this;
  };

  flow.margin =
  function( _ )
  {
    if ( !arguments.length ) return margin;
    flow.top     = typeof _.top    != 'undefined'  ? _.top     : flow.top;
    flow.right   = typeof _.right  != 'undefined'  ? _.right   : flow.right;
    flow.bottom  = typeof _.bottom != 'undefined'  ? _.bottom  : flow.bottom;
    flow.left    = typeof _.left   != 'undefined'  ? _.left    : flow.left;
    return this;
  };

  flow.width =
  function( value )
  {
    if ( !arguments.length ) return width;
    width = parseInt( value );
    return this;
  };

  flow.height =
  function( value )
  {
    if ( !arguments.length ) return height;
    height = parseInt( value );
    return this;
  };

  flow.nodes = 
  function( value )
  {
    if ( !arguments.length ) return nodes;
    nodes = value;
    return this;
  };

  flow.links = 
  function( value )
  {
    if ( !arguments.length ) return links;
    links = value;
    return this;
  };

  // TODO Unused functions - start
    flow.nodeWidth = function(value) {
        if (!arguments.length) return nodeWidth;
        nodeWidth = parseInt(value);
        return this;
    };

    flow.nodeHeight = function(value) {
        if (!arguments.length) return nodeHeight;
        nodeHeight = parseInt(value);
        return this;
    };

    flow.containerHeight = function(value) {
        if (!arguments.length) return containerHeight;
        containerHeight = parseInt(value);
        return this;
    };
  // TODO Unused functions - end 

  return flow;
}
