/// Global variables

// map< fileinfo.index, position of fileinfo in fileInfos array (which is
// the same as the tab index)>
// i.e. map< file index, tab index >
var tabIndexMap = {};

// vector< fileInfo objects >
var fileInfos;

var disableClusterEdgesFlag = false;

function
main()
{
  var activeKernel = 0;

  var mavData;
  if ( DEBUG ) mavData = JSON.parse( mavJSON );
  else         mavData = JSON.parse( sci.getMAVJSON() );

  // FIXME temporary flag for disabling edge clustering
  if ( !DEBUG ) {
    var flagString = sci.getClusterEdgesFlag();
    if ( flagString === "true" )
      disableClusterEdgesFlag = true;
  }

  // map< file name, file index > (The map used by the compiler)
  var fileIndexMap = mavData.fileIndexMap;
  
  // 1. Add kernel names to the dropdown menu
  addKernelNames( mavData.functions );

  // 2. Gather file names and content in one structure
  fileInfos = parseFileInfos();
  tabIndexMap = createTabIndexMap( fileInfos );

  // 3. For each source file, add a tab in the editor pane
  addFileTabs( fileInfos );
  
  // 4. Now that we have all tab bars, resize the main panes to fit the
  // window height 
  resizeColumns();

  // 4. Add file contents
  addFileContents( fileInfos );

  // 5. Add onclick functions to report tabs (they're already statically added
  // in index.html) 
  addReportTabs();

  verifyThings();

  // 6. Draw the active report tab
  drawActiveReportTab( activeKernel );

  ///// Functions
  
  /// main::resizeColumns
  function
  resizeColumns()
  {
    $( window ).on( "resize", resizeToHeight ); 
    resizeToHeight();
  }

  /// main::resizeToHeight
  function
  resizeToHeight()
  {
    var windowHeight = $( window ).height();
    // Get the height including paddings/margins
    var headerBarHeight =  $( "#header-bar" ).outerHeight( true );
    // Allow some margin below 
    var bottomMargin = 20;
    
    var tabBarHeight = $( "#graph-pane .nav-tabs" ).outerHeight( true );
    var newHeight = windowHeight - headerBarHeight - tabBarHeight - bottomMargin;
    assert( newHeight > 0 );
    $( "#graph-pane" ).height( newHeight );
    
    // Each tab bar may have a different height
    tabBarHeight = $( "#editor-pane .nav-tabs" ).outerHeight( true );
    newHeight = windowHeight - headerBarHeight - tabBarHeight - bottomMargin;
    assert( newHeight > 0 );
    $( "#editor-pane" ).height( newHeight );

    tabBarHeight = $( "#details-pane .nav-tabs" ).outerHeight( true );
    newHeight = windowHeight - headerBarHeight - tabBarHeight - bottomMargin;
    assert( newHeight > 0 );
    $( "#details-pane" ).height( newHeight );

  }

  /// main::parseFileInfos
  function
  parseFileInfos()
  {
    var fileInfos = [];
  
    if ( DEBUG ) {
      fileInfos = JSON.parse( fileJSON ); 
  
      // Replace the file info indices with those from the fileIndexMap
      fileInfos.forEach( function( d ) {
        var index = fileIndexMap[ d.path ];
        assert( index > 0 ); 
        d.index = index;
      } );
    }
    else {
      for ( var i = 0, e = sci.getNumFiles(); i != e; ++i ) {
        var filePath = sci.getFullFileName( i );
        var index = fileIndexMap[ filePath ];
        assert( index > 0 );
        
        var info = {};
        info.index = index;
        info.path = filePath;
        info.name = sci.getFileName( i );
        info.content = sci.getFileContents( i ); 
        
        fileInfos.push( info );
      }
    }
    
    verifyFileInfos( fileInfos );
  
    return fileInfos;
  }
  
  /// main::verifyFileInfos
  function
  verifyFileInfos( fileInfos )
  {
    fileInfos.forEach( function( d ) {
      assert( d.index == fileIndexMap[ d.path ] );
    } );
  }
  
  /// main::createTabIndexMap
  function
  createTabIndexMap( fileInfos )
  {
    var tabIndexMap = {};
    fileInfos.forEach( function( d, i ) {
      tabIndexMap[ d.index ] = i;
    } );
  
    return tabIndexMap;
  }
  
  /// main::addKernelNames
  function
  addKernelNames()
  {
    var names = [];
    mavData.functions.forEach( function( d ) {
      names.push( d.name );
    } );
  
    var navBar = d3.select( "#header-bar" ).select( ".navbar-nav" );
    var dropdown = navBar.select( ".dropdown-menu" );
    var listElements = dropdown.selectAll( "li" )
      .data( names )
      .enter()
      .append( "li" );
  
    var anchors = listElements
      .append( "a" )
      .attr( "href", "#" )
      .text( function( d ) { return d } )
      .on( "click", function( d, i ) {
        d3.event.preventDefault();
        activeKernel = i;
        drawActiveReportTab( i );
      } );
  }
  
  /// main::findActiveReportTab
  function
  findActiveReportTab()
  {
    var navTabs = d3.select( "#graph-pane" ).select( ".nav-tabs" );
    var activeTab = navTabs.selectAll( "li.active" );
    var tabIndex = $( activeTab.node() ).index();
    // Only 2 reports for now
    assert( tabIndex < 2 );
  
    return tabIndex;
  }
  
  /// main::drawActiveReportTab
  function
  drawActiveReportTab( kernelIndex )
  {
    // Call syncPanes with empty object to clear the details pane
    syncPanesToNode( {} );

    var activeTab = findActiveReportTab();
    drawTab( activeTab, kernelIndex );
  }
  
  /// main::drawTab
  function
  drawTab( tabIndex, kernelIndex )
  {
    switch ( tabIndex ) {
      case 0:
        {
          var graphRoot = mavData.functions[ kernelIndex ];
          startGraph( graphRoot );
        }
        break;
      default:
        assert( false );
        break;
    }
  }
  
  /// main::addReportTabs
  function
  addReportTabs()
  {
    // Any time a tab is shown, update the contents
    $( document ).on( 'shown.bs.tab', 'a[data-toggle="tab"]', function( e ) {
      var anchor = e.target;
      
      // If the anchor is a descendant of the graph-pane nav tab
      if ( $( "#graph-pane .nav-tabs" ).find( anchor ).length > 0 ) {
        drawActiveReportTab( activeKernel );
      }
    } );
  }

  /// main::verifyThings
  function
  verifyThings()
  {
    // 1. Verify fileIndex/tabIndex maps:
    for( var filename in fileIndexMap ) {
      if ( !fileIndexMap.hasOwnProperty( filename ) ) continue;

      var fileIndex = fileIndexMap[ filename ];
      var tabIndex = tabIndexMap[ fileIndex ];
      assert( tabIndex === parseInt( tabIndex ) ); // Ensure is integer

      // Get the tab at that index
      var theTab = $( "#editor-pane .nav-tabs" ).children().eq( tabIndex );
      // Now check that the filename (full file path) ends with the tab name
      assert( filename.endsWith( theTab.text() ) );
    }

    // 2. Verify the order of the dropdown menu items matches the JSON data
    // Get the dropdown menu
    var dropdownAnchors = $( "#header-bar .navbar-nav .dropdown-menu li a" );
    dropdownAnchors.each( function( index ) {
      assert( mavData.functions[ index ].name === $( this ).text() );
    } );
    
  }
}

// TODO Any functions that are called only once should be moved to the
// callee body

function
addFileTabs( fileInfos )
{
  var navTabs = d3.select( "#editor-pane" ).selectAll( ".nav-tabs" );
  var listElements = navTabs.selectAll( "li" )
    .data( fileInfos )
    .enter()
    .append( "li" )
    .attr( "class", function( d, i ) {
        var classname = "";
        if ( i == 0 ) classname =  "active ";
        return classname;
      } );

// TODO add data-toggle=tab to these
  var anchors = listElements  
    .append( "a" )
    .attr( "href", function( d ) { return "#file" + d.index; } )
    .attr( "role", "tab" )
    .text( function( d ) { return d.name; } )
    ;

  $( "#editor-pane-nav" ) 
    .on( "click", "a", function( e ) {
      e.preventDefault();
      $( this ).tab( "show" );
    } );
}

function
addFileContents( fileInfos )
{
  var tabContent = d3.select( "#editor-pane" ).selectAll( ".tab-content" );

  var divs = tabContent.selectAll( "div" )
    .data( fileInfos )
    .enter()
    .append( "div" )
    .attr( "class", function( d, i ) {
      var classname = "tab-pane fade";
      if ( i == 0 ) classname = classname + " in active";
      return classname;
    } )
    .attr( "id", function( d ) { return "file" + d.index; } )
    ;
  var editorDivs = divs
    .append( "div" )
    .attr( "class", "well" );

  editorDivs.each( setupEditor );

  /// Functions
  function
  setupEditor( fileInfo )
  {
    var editor = ace.edit( this ); // "this" is the DOM element
    fileInfo.editor = editor;
    
    // TODO cleanup relative paths 
    editor.setTheme( "../ace/theme/xcode" );
    editor.setFontSize( 12 );
    editor.getSession().setMode( "../ace/mode/c_cpp" );
    editor.getSession().setUseWrapMode( true );  
 
    // Replace \r\n with \n in the file content (for windows)
    editor.setValue( fileInfo.content.replace( /(\r\n)/gm, "\n" ) );
    editor.setReadOnly( true );
    editor.scrollToLine( 1, true, true, function() {} );
    editor.gotoLine( 1 );
 }
}

///// Global functions

/// Syncs the editor and details pane to this node
function
syncPanesToNode( node )
{
  syncDetailsPaneToNode( node );
  syncEditorPaneToNode( node );
}

// Assumes the node has the file index and line number 
// in "file" and "line" respectively
function
syncEditorPaneToNode( node )
{
  // Note: This check returns true if node.file is undefined or if it's  0
  // This is good because file index 0 is used for unknown
  if ( !node.file ) return;

  var tabIndex = tabIndexMap[ node.file ];
  var target = "li:eq(" + tabIndex + ")";
  $( "#editor-pane-nav " + target + " a" ).tab( "show" ); 

  var editor = fileInfos[ tabIndex ].editor;
  assert( editor );
  var line = node.line;
  assert( line > 0 );
  editor.focus();
  editor.resize( true );
  editor.scrollToLine( line, true, true, function() {} );
  editor.gotoLine( line );
}

function
syncDetailsPaneToNode( node )
{
  // Clear contents
  var tbody = d3.select( "#details-pane table tbody" );
  tbody.each( function() { while( this.lastChild ) this.removeChild( this.lastChild ); } );

  if ( !node.details ) return;
  
  var rows = tbody 
    .selectAll( "tr" )
    .data( node.details )
    .enter()
    .append( "tr" );

  rows.append( "td" )
    .text( function( d ) { return d.first; } )
    .style( "word-wrap", "break-word" );
  rows.append( "td" )
    .text( function( d ) { return d.second; } )
    .style( "word-wrap", "break-word" );
}

// For calling endsWith on strings
if (!String.prototype.endsWith) {
  Object.defineProperty(String.prototype, 'endsWith', {
    value: function (searchString, position) {
      var subjectString = this.toString();
      if (position === undefined || position > subjectString.length) {
        position = subjectString.length;
      }
      position -= searchString.length;
      var lastIndex = subjectString.indexOf(searchString, position);
      return lastIndex !== -1 && lastIndex === position;
    }
  });
}
