#include "host/inc/decode_types.h"
#include "host/inc/msg_types.h"

// Input interface
channel ulong2 input_stream __attribute__((depth(0))); // 8 bytes / cycle from UDP channel + control

// Test output interface

// Array of channels dumping messages
channel ulong output_stream[FIELDS] __attribute__((depth(0))); 

// Utility functions processing UDP channel control info

bool is_eop(ulong ctrl_data) {
   bool ret = (ctrl_data & EOP_MASK) >> EOP_BIT ? true : false;
   return ret;
}

uchar extract_valid_bytes(ulong ctrl_data) {
   uchar a = (ctrl_data & VALID_MASK) >> VALID0_BIT;
   return a;
}

// Utility function counting the number of 1s in a ulong

unsigned char cl_count_ones(unsigned long bits) {
    unsigned long count = bits;
    count = ( count & 0x5555555555555555 ) + ( ( count >> 1 ) & 0x5555555555555555 ) ;
    count = ( count & 0x3333333333333333 ) + ( ( count >> 2 ) & 0x3333333333333333 );
    count = ( count & 0x0f0f0f0f0f0f0f0f ) + ( ( count >> 4 ) & 0x0f0f0f0f0f0f0f0f );
    count = ( count & 0x00ff00ff00ff00ff ) + ( ( count >> 8 ) & 0x00ff00ff00ff00ff );
    count = ( count & 0x0000ffff0000ffff ) + ( ( count >> 16 ) & 0x0000ffff0000ffff );
    count = ( count & 0x00000000ffffffff ) + ( ( count >> 32 ) & 0x00000000ffffffff );
    return (char)count;
}

// Utility function returning a mask matching the nth 1 in an ulong, the count starts from 0
ulong findNthPop(ulong pmap, int n) {
   ulong mask = 1;
   ulong result = 0;
   bool found = false;
   #pragma unroll 64
   for (int i = 0; i < 64; i++) {
      if (!found && cl_count_ones(pmap & mask) == n + 1) {
         result |= 1 << i;
         found = true;
      }
      mask <<= 1;
      mask |= 1;
   }
   return result;
}

// Utility reversing 7 * n bits in an ulong (n is encoded as bit-vector marking
// a bit as 1 for each group of 7 bits)
ulong reverse(ulong x, uchar n) {
   ulong result = 0;
   #pragma unroll 64
   for (int i = 0; i < 64; i++) {
      if ((n >> (i / 7)) & 0x01) {
         result <<= 1; 
         result |= x & 1;
      }
      x >>= 1;
   }
   return result;
}

// OPRA decoder kernel

__attribute__((task))
__attribute__((autorun))
kernel void OPRADecoder() {
   uchar left = OPRA_HEADER_SIZE; // keeps track of bytes left in current message
   bool valid = false; // current message is th OPRA header, so not valid

   ulong crt_field = 0; // carries a partially parsed field
   uchar crt_length = 0; // bit-vector marking the number of bytes received for crt_field

   ulong pmap = 0; // carries current message's pmap
   uchar pmapLength; // carries the length of the pmap
   bool pmapParsed = false; // flags whether the pmap of current message was parsed

   uchar fieldCount = 0; // counts fields parsed in the current message

   uchar8 stream_data; // UDP data - 8 bytes / word
   ulong ctrl_data;  // ctrl word associated with data word
   ulong long_data;

   ulong msg[FIELDS]; // message data

   bool persist = false; // flags whether the new iteration should reprocess previous word

   while(get_global_id(0)==0) { 
      // read a word, unless the previous word was not fully used
      if (!persist) {
         ulong2 data = read_channel_altera(input_stream);
         ctrl_data = data.y;
         long_data = data.x;
         stream_data = *((uchar8 *)&(long_data)); // cast to uchar8 for easy access to elements
      }
      
      ulong completedMessage[FIELDS];
      #pragma unroll FIELDS
      for (int i = 0; i < FIELDS; i++) {
         completedMessage[i] = i+long_data+ctrl_data;
      }


      #define WRITE(i) write_channel_altera(output_stream[i], completedMessage[i]);

      if (1) {
         WRITE(0); WRITE(1); WRITE(2); WRITE(3); WRITE(4); 
         WRITE(5); WRITE(6); WRITE(7); WRITE(8); WRITE(9); 
         WRITE(10); WRITE(11); WRITE(12); WRITE(13); WRITE(14); 
         WRITE(15); WRITE(16); WRITE(17); WRITE(18); WRITE(19); 
         WRITE(20); WRITE(21); WRITE(22); WRITE(23); WRITE(24); 
         WRITE(25); WRITE(26); WRITE(27); WRITE(28); WRITE(29); 
         WRITE(30); WRITE(31); WRITE(32); WRITE(33); WRITE(34); 
         WRITE(35); WRITE(36); WRITE(37); WRITE(38); WRITE(39); 
         WRITE(40); WRITE(41); WRITE(42); WRITE(43); WRITE(44); 
         WRITE(45); WRITE(46); WRITE(47); WRITE(48); WRITE(49); 
      }

   }
}


