d3.custom.chart.flow = function() {

    // public variables with default settings
    var //margin = {top:10, right:10, bottom:10, left:10}, // defaults
        margin = {top:15, right:15, bottom:15, left:15}, // defaults
        //padding = {top:20, right:10, bottom:10, left:10},
        padding = {top:25, right:15, bottom:15, left:15},
        transitionDuration = 300,
        chartGroup,
        container,
        svg,
        width,
        height,
        root,
        rootNode,
        scrollbarAffordance;

    var flow = d3.custom.layout.flow()
        .margin(margin)
        .padding(padding)
        .nodeWidth(110)
        .nodeHeight(30)
        .containerHeight(20);
     
    function chart(selection) {
        rootNode = selection.node();

        function debounce(fn, timeout) {
            var timeoutID = -1;
            return function() {
                if (timeoutID > -1) {
                    window.clearTimeout(timeoutID);
                }
                timeoutID = window.setTimeout(fn, timeout);
            }
        }

        function resize(selectedNode) {
            var domContainerWidth  = (parseInt(d3.select(rootNode).style("width"))),
                domContainerHeight = (parseInt(d3.select(rootNode).style("height"))),
                flowWidth = 0;

            if (root.height > domContainerHeight) {
                scrollbarAffordance = 0;
            } else {
                scrollbarAffordance = 0;
            }

            flowWidth = domContainerWidth - scrollbarAffordance;
            flow.width(flowWidth);

            chart.update(selectedNode);

            svg.transition().duration(transitionDuration)
                .attr("width", function(d) {
                    return domContainerWidth;
                })
                .attr("height", function(d) {
                    return d.height + margin.top + margin.bottom;
                })
                .select(".chartGroup")
                .attr("width", function(d) {
                    return flowWidth;
                })
                .attr("height", function(d) {
                    return d.height + margin.top + margin.bottom;
                })
                .select(".background")
                .attr("width", function(d) {
                    return flowWidth;
                })
                .attr("height", function(d) {
                    return d.height + margin.top + margin.bottom;
                });
        }


        d3.select(window).on('resize', function() {
            debounce(resize, 50)();
        });


        selection.each(function(arg) {
            root = arg;
            container = d3.select(this);

            var i = 0;

            if (!svg) {
                svg = container.append("svg")
                    .attr("class", "svg chartSVG")
                    .attr("transform", "translate(0, 0)")
                    .style("shape-rendering", "auto") // shapeRendering options; [crispEdges|geometricPrecision|optimizeSpeed|auto]
                    .style("text-rendering", "auto"); // textRendering options;  [auto|optimizeSpeed|optimizeLegibility|geometricPrecision]
                chartGroup = svg.append("svg:g")
                    .attr("class", "chartGroup");
                chartGroup.append("svg:rect")
                    .attr("class", "background");
            }

            chart.update = function(source) {
                flow(root);
                var nodes = flow.nodes(),
                    links = flow.links();

                function color(d) {
//                    return d._children ? "gold" : d.children ? "white" : "aquamarine";
                    return d._children ? "lightgray" : d.children ? "white" : "white";
                }

                // Toggle children on click.
                function click( node ) {
                  switch( node.exp ) {
                    case ExpEnum.ONE:
                      {
                        // We are expanding this node
                        if ( node._children ) {
                          var parent = node.parent;
                          if ( node.parent ) {
                            // Only visible siblings can be in expanded view 
                            var siblings = node.parent.children;
                            assert( siblings, "parent should have visible children" ); 

                            // collapse other siblings
                            siblings.forEach( function( sibling ) {
                                if ( sibling.children ) {
                                // TODO test this
                                // Make sure there are no hidden children
                                assert( !sibling._children, "node has both visible and hidden children." );

                                sibling._children = sibling.children;
                                sibling.children = null;
                                }
                                } );

                            // Make this node's children visible
                            assert( !node.children, "node should not have visible children at this point." );
                            node.children = node._children;
                            node._children = null;
                          }
                        }
                        // We are collapsing the node
                        else {
                          node._children = node.children;
                          node.children = null;
                        }
                      }
                      break;
                    
                    case ExpEnum.PASS:
                      {
                        var parent = node.parent;
                        if ( parent ) click( parent );
                      }
                      break;

                    case ExpEnum.ALL:
                      {
                        if ( node.children ) {
                          node._children = node.children;
                          node.children = null;
                        } else {
                          node.children = node._children;
                          node._children = null;
                        } 
                      }
                      break;
                    
                    case ExpEnum.OFF:
                    default:
                      {
                        // Do nothing
                      }
                      break;

                  }
                  // FIXME should be resizing the node or its parent?
                  resize( node );
                }

                function
                dblclick( node )
                {
                  if ( node.line ) editor.gotoLine( node.line );
                }

                // Update the nodes
                var node = chartGroup.selectAll("g.node")
                    .data(nodes, function(d) { return d.id || (d.id = ++i); });

                var nodeEnter = node.enter().append("svg:g")
                    .attr("class", function( d ) {
                      return "node " + d.shape;
                    } )
                    .attr("transform", function(d) {
                        return "translate(" + source.x + "," + source.y + ")";
                    })
                    .style("opacity", 1e-6);
                
                var rectEnter = nodeEnter.filter( ".RECT" );    
                var circEnter = nodeEnter.filter( ".CIRC" );

                // Enter any new nodes at the parent's previous position.
                rectEnter.append("svg:rect")
                    .attr("class", "background")
                    .attr("height", function(d) { return d.height; })
                    .attr("width", function(d) { return d.width; })
                    .style("fill", color)
                    .on("click", click)
                    .on("dblclick", dblclick);
                
                circEnter.append("svg:circle")
                    .attr("r", function(d) { return d.width/2; })
//                    .attr( "cx", function(d) { return d.x; } ) 
//                    .attr( "cy", function(d) { return d.y; } ) 
                    .style("fill", color)
                    .on("click", click)
                    .on("dblclick", dblclick);
                
                nodeEnter.each(function(d) {
//                    if (d.children || d._children) {
//                        d3.select(this)
//                            .append("path")
//                            .attr("class", "expander")
//                            .attr("d", "M 0 0 L 6 6 L 0 6 z")
//                            .attr("transform", function(d) {
//                                return d._children ? "translate(8,14)rotate(225)" : "translate(5,8)rotate(315)";
//                            });
//                        d3.select(this).append("svg:text")
//                            .attr("class", "label")
//                            .attr("dy", 13)
//                            .attr("dx", 17)
//                            .text(function(d) { return d.name; });
//                    } else {
                        d3.select(this).append("svg:text")
                            .attr("class", "label")
                            .attr("dy", 10)
                            .attr("dx", 10)
                            .text(function(d) { 
                              if ( d.type == "port" ) return "";
                              return d.name; 
                            });
//                    }
                });

                // Transition nodes to their new position.
                nodeEnter.transition()
                    .duration(transitionDuration)
                    .attr("transform", function(d) { return "translate(" + d.x + "," + d.y + ")"; })
                    .style("opacity", 1);

                var nodeUpdate = node.transition()
                    .duration(transitionDuration);

                var rectUpdate = nodeUpdate.filter( ".RECT" );
                var circUpdate = nodeUpdate.filter( ".CIRC" );

                rectUpdate.attr("transform", function(d) { return "translate(" + d.x + "," + d.y + ")"; })
                    .style("opacity", 1)
                    .select("rect")
                    .style("fill", color);

                circUpdate.attr("transform", function(d) { return "translate(" + d.x + "," + d.y + ")"; })
                    .style("opacity", 1)
                    .select("circle")
                    .style("fill", color);
                
//                nodeUpdate.each(function(d) {
//                    if (d.children || d._children) {
//                        d3.select(this).select(".expander").transition()
//                            .duration(transitionDuration)
//                            .attr("transform", function(d) {
//                                return d._children ? "translate(8,14)rotate(225)" : "translate(5,8)rotate(315)";
//                            });
//                    }
//                });
//
                // assume only rectangular nodes can be background
                nodeUpdate.select(".background")
                    .attr("height", function(d) { return d.height; })
                    .attr("width", function(d) { return d.width; });

                // Transition exiting nodes to the parent's new position.
                node.exit().transition()
                    .duration(transitionDuration)
                    .attr("transform", function(d) { return "translate(" + source.x + "," + source.y + ")"; })
                    .style("opacity", 1e-6)
                    .remove();


               // Update the links
               var link = svg.selectAll( "path.link" )
//                              .data( links, function( d ) { return d.target.id; } );
                              .data( links );

              var line = d3.svg.line()
                                .x( function( d ) { return d.x; } )
                                .y( function( d ) { return d.y; } );

              function
              drawLine( link )
              {
//                console.log( "original link: " );
//                console.log( link );

                var source = {}, target = {};

                switch( link.source.shape ) {
                  case "CIRC":
                    {
                      source.x = link.source.x - link.source.width/2;
                      source.y = link.source.y;
                    }
                    break;
                  
                  case "RECT":
                    {
                      source.x = link.source.x + link.source.width;
                      source.y = link.source.y + link.source.height/2;
                    }
                    break;

                  default:
                    {
                      source.x = link.source.x;
                      source.y = link.source.y;
                    }
                }
                
                switch( link.target.shape ) {
                  case "CIRC":
                    {
                      target.x = link.target.x - link.target.width/2;
                      target.y = link.target.y;
                    }
                    break;
                  
                  case "RECT":
                    {
                      target.x = link.target.x + link.target.width;
                      target.y = link.target.y + link.target.height/2;
                    }
                    break;

                  default:
                    {
                      target.x = link.target.x;
                      target.y = link.target.y;
                    }
                }

                var points = [ source, target ]; 
//                console.log( "points" );
//                console.log( points );
                return line( points ); 
              }

              // TODO rename
              function
              fix( node )
              {
                switch( node.shape ) {
                  case "CIRC":
                    return [ node.x - node.width/2, node.y ]
                  
                  case "RECT":
                    return [ node.x + node.width/2, node.y ]

                  default:
                    return [ node.x, node.y ];
                }
              } 

              var diagonal = d3.svg.diagonal()
                                    .projection( function( d ) {
//                                      return [ d.x, d.y ];
                                      return fix( d );
                                     } );

               // Enter any new links at the parent's pervious position.
               // TODO figure out how to group all links together at the top
               // in DOM
               link.enter().append( "path" )
                           //.insert( "path", "g" )
                            .attr( "class", "link" )
                            .attr( "d", function( d ) {
                              // FIXME this should be source.x0, y0
                              var o = { x: source.x, y: source.y };
                              return diagonal( { source:o, target:o } );
                             } );

               // Transition links to their new position.
               link.transition()
                    .duration( transitionDuration )
                    .attr( "d", diagonal );

               // Transition exiting nodes to the parent's new position.
               link.exit().transition()
                            .duration( transitionDuration )
                            .attr( "d", function( d ) {
                              var o = { x:source.x, y:source.y };
                              return diagonal( { source:o, target:o } );
                            } )
                            .remove();
/*

               // Enter any new links at the parent's pervious position.
               // TODO figure out how to group all links together at the top
               // in DOM
               link.enter().append( "path" )
                           //.insert( "path", "g" )
                            .attr( "class", "link" )
                            .attr( "d", function( d ) {
                              var link = { source:{}, target:{} };
//                              link.source.x = d.source.x0;
//                              link.source.y = d.source.y0;
//                              link.source.width = d.source.width;
//                              link.source.height = d.source.height;
//                              
//                              link.target.x = d.target.x0;
//                              link.target.y = d.target.y0;
//                              link.target.width = d.target.width;
//                              link.target.height = d.target.height;

                              return drawLine( d );
                             } );

               // Transition links to their new position.
               link.transition()
                    .duration( transitionDuration )
                    .attr( "d", drawLine );

               // Transition exiting nodes to the parent's new position.
               link.exit().transition()
                            .duration( transitionDuration )
                            .attr( "d", drawLine )
                            .remove();
*/               

                // FIXME these are currenlty not used
                // Remember old positions for transition
                nodes.forEach( function( d ) {
                  d.x0 = d.x;
                  d.y0 = d.y;
                 } );
            };

            resize(root);
            
            // FIXME not currently used
            root.x0 = 0, root.y0 = 0;
            chart.update(root);

        });
    }

    chart.width = function(value) {
        if (!arguments.length) return width;
        width = parseInt(value);
        return this;
    };

    chart.height = function(value) {
        if (!arguments.length) return height;
        height = parseInt(value);
        return this;
    };

    chart.margin = function(_) {
        if (!arguments.length) return margin;
        margin.top    = typeof _.top    != 'undefined' ? _.top    : margin.top;
        margin.right  = typeof _.right  != 'undefined' ? _.right  : margin.right;
        margin.bottom = typeof _.bottom != 'undefined' ? _.bottom : margin.bottom;
        margin.left   = typeof _.left   != 'undefined' ? _.left   : margin.left;
        return chart;
    };

    return chart;
};
