    
// Global variables
var nodeIdMap = {};

// OFF: the node itself cannot be expanded
// ONE: only one sibling can be expanded at a time
// PASS: the node itself doesn't respong to clicks but passes it to parent
// ALL: default, everything can be expanded 
var ExpEnum = { OFF: "off", ONE: "one", PASS: "pass", ALL: "all" };

window.onload = start( mavData );
window.onresize = start( mavData );

// TODO remove all references to kernel node

// Main function
function
start( mavData )
{
  jsonData = JSON.parse( mavData );
  data = jsonData.functions[ 0 ];

  // 1. Prep the data in various ways
  // 1a. Create a map of node id -> node object
  // This will be used to create the link objects
  nodeIdMap = populateNodeIdMap( data );

  // 1b. Add shape-related attribute to each node based on its type
  addShapeAttributes( data );

  // 1c. For each node, decide how many of the children can be expanded at
  // a time 
  addExpandAttribute( data );

  // 1d. For each node, decide whether children should be hidden or visible
  // by default
  addChildrenVisibility( data );

  // 1e. For each node, add a pointer to its parent
  // This is used to collapse one child if another is expanded for nodes that
  // require this functionality
  addParent( data );
  
  // 2. Create the diagram
  var flowTree = d3.custom.chart.flow();

  d3.select('#flowtree')
    .datum(data)
    .call(flowTree)
    .on("mouseover.customHover", function() {
        //    console.log("customHover");
     } )
    .on("resize.custom", function() {
      //    console.log("resize.custom");
     } );
}

function
populateNodeIdMap( root )
{
  var stack = [ root ],
      node,
      nodeMap = {};

  while( ( node = stack.pop() ) != null ) {
    nodeMap[ node.id2 ] = node;
    
    var children = node._children || node.children;
    if ( !children ) continue;
    
    var n = children.length;
    while ( --n >= 0 ) {
      stack.push( children[ n ] );
    }
  }

  return nodeMap;
}

function
addShapeAttributes( node )
{
  // Set the node shape and intrinsic width/height
  switch( node.type ) {
    case "merge": 
    case "branch":
    case "port":
      {
        node.shape = "CIRC";
        // TODO fix these to be 20 once I've moved them
        // to the basic block borders
        // In the case of the circle iw=ih=diameter
        node.iw = 8;
        node.ih = 8;
      }
      break;

    case "kernel":
    case "container":
      {
        node.shape = "RECT";
        node.iw = 120;
        node.ih = 60;
      }
      break;

    case "bb":
    case "memtype":
    case "memsys":
    case "bank":
      {
        node.shape = "RECT";
        node.iw = 100;
        node.ih = 40;
      }
      break;

    case "inst":
      {
        node.shape = "RECT";
        node.iw = 80;
        node.ih = 20;
      }
      break;

    default:
      // TODO
      break;
  }

  var children = node._children ? node._children : node.children;
  if ( !children ) return;
  children.forEach( addShapeAttributes );
}

// OFF: the node itself cannot be expanded
// ONE: only one sibling can be expanded at a time
// PASS: the node itself doesn't respong to clicks but passes it to parent
// ALL: default, everything can be expanded 
// var ExpEnum = { OFF: "off", ONE: "one", PASS: "pass", ALL: "all" };
function
addExpandAttribute( node )
{
  // Set node.exp to one of the above expand options 
  switch( node.type ) {
    case "memtype":
    case "memsys":
      node.exp = ExpEnum.ONE;
      break;

    case "bank":
      node.exp = ExpEnum.PASS;
      break;

    case "inst":
    case "port":
      node.exp = ExpEnum.OFF;
      break;

    default:
      node.exp = ExpEnum.ALL;
      break;
  }

  var children = node._children ? node._children : node.children;
  if ( !children ) return;
  children.forEach( addExpandAttribute );
}

function
addChildrenVisibility( node )
{
  assert( !node._children );

  switch( node.type ) {
    // Visible by default 
    case "container":
//    case "bb":
    case "bank":
      // Do nothing
      break;

    // Hidden by default
    default:
      {
        node._children = node.children;
        node.children = null;
      } 
      break;
  }

  var children = node._children ? node._children : node.children;
  if ( !children ) return;
  children.forEach( addChildrenVisibility );
}

function
addParent( node )
{
  // If the parent hasn't already been set, set it to null (for the root)
  if ( !node.parent ) node.parent = null;
  
  var children = node._children ? node._children : node.children;
  if ( !children ) return;
  
  children.forEach( function( child ) {
    child.parent = node;
    addParent( child );
  } );
}


