d3.custom.layout.flow = 
function() 
{
  var hierarchy = d3.layout.hierarchy().sort( null ).value( null ),
      width     = 0,
      height    = 0,
      padding   = { top:20, left:10, bottom:10, right:10 },
      margin    = { top:10, left:10, bottom:10, right:10 };

  var nodes = [],
      links = [];
  
  // TODO Unused variables
  var nodeWidth = 0,
      nodeHeight = 0,
      containerHeight = 0;
  
  // Replace all uses of nodeWidth and nodeHeight with node.iw and node.ih
  // containerWidth and containerHeight have to be calculated based on whether
  // the nodes is showing its children or not

  function 
  flow( d, i )
  {
    nodes = hierarchy.call( this, d, i );
    var root  = nodes[ 0 ];

//    console.log( "flow nodes:" );
//    console.log( nodes );

    function
    gatherLinks( nodes )
    {
      return d3.merge( nodes.map( function( node ) {
        return ( node.edges || [] ).map( function( edge ) {
          // FIXME should we pass nodeIdMap to the constructor?
          return { source: nodeIdMap[ edge.from ], target: nodeIdMap[ edge.to ] };
        } )
      } ) );
    }

    function
    firstWalk( node )
    {
      var children = node.children;
      // The node has visible children
      if ( children && children.length > 0 ) {
        children.forEach( firstWalk );
        layoutNode( node );
      }
      else {
        node.x = 0;
        node.y = 0;
        node.width = node.iw;
        node.height = node.ih;
      }
    }

    function
    secondWalk( node )
    {
      var children = node.children;
      if ( !children ) return;
      children.forEach( function( child ) {
        child.x += node.x;
        child.y += node.y;
        secondWalk( child );
      } );
    }

    // Layout a node and its children
    // This function should only be called on nodes that have (visible)
    // children
    function
    layoutNode( node )
    {
      // Lay out things differently based on the node type
      switch( node.type ) {
        case "memsys":
          layoutMemSystem( node );
          break;

        case "bank":
          layoutBank( node );
          break;

        default:
          gridLayout( node );
          break;
      }
    }

    function
    layoutMemSystem( node )
    {
      var   children  = node.children
          , curX      = 0     
          , curY      = 0     
          , width     = 0     
          ;

      children.forEach( function( child ) {
        child.x = curX;
        child.y = curY;
      
        curY += child.height;

        // Ensure all children have the same width
        if ( width == 0 ) width = child.width;
        else              assert( width == child.width );
      } );    

      node.width = width;
      node.height = curY;
      
      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }

    function
    layoutBank( node )
    {
      var   children  = node.children
          , curY      = padding.top/2 // TODO 
          ;
      
      children.forEach( function( child ) {
        child.x = 0;
        child.y = curY + child.height/2;

        curY += child.height/2 + padding.bottom;

        // Ensure that the child has a smaller width than this node
        assert( child.width < node.iw );
      } );

      node.width = node.iw;
      node.height = Math.max( curY, node.ih ); 

      node.x = 0;
      node.y = 0;
    }

    function
    gridLayout( node )
    {
      var   children  = node.children
          , curX      = 0 
          , curY      = 0 
          , maxWidth  = 0
          , maxHeight = 0
          , prev      = null;
      
      children.forEach( function( child ) {
        if ( prev ) {
          if ( prev.order == child.order ) {
            // same row, next column
            curX += prev.width + padding.right;
          }
          else {
            // next row, column 0
            curY += prev.height + padding.bottom;
            curX = 0;
          }
        }

        // (curX, curY) is the starting point of the cell
        // Move to the middle of the cell
        child.x = curX + padding.left;
        child.y = curY + padding.top;
        
        maxWidth = Math.max( maxWidth, child.x + child.width + padding.right );
        maxHeight = Math.max( maxHeight, child.y + child.height + padding.bottom );
        prev = child;
      } );

      node.width = maxWidth;
      node.height = maxHeight;

      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }

    // Layout a node and its children
    // This function should only be called on nodes that have (visible)
    // children
    function
    gridLayout0( node )
    {
      var   children  = node.children
          , curX      = padding.left
          , curY      = padding.top
          , maxWidth  = 0;
      
      children.forEach( function( child ) {
        child.x = curX;
        child.y = curY;

        curY += child.height + padding.bottom;
        maxWidth = Math.max( maxWidth, child.width + padding.right );
      } );

      node.width = curX + maxWidth;
      node.height = curY;
      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }

    firstWalk( root );
    secondWalk( root );
    width = root.width;
    height = root.height; 

    links = gatherLinks( nodes );
//    console.log( "links: " );
//    console.log( links );

    return nodes;
  }

  // TODO understand what these are doing!
  flow.padding =
  function( _ )
  {
    if ( !arguments.length ) return padding;
    padding.top     = typeof _.top    != 'undefined'  ? _.top     : padding.top;
    padding.right   = typeof _.right  != 'undefined'  ? _.right   : padding.right;
    padding.bottom  = typeof _.bottom != 'undefined'  ? _.bottom  : padding.bottom;
    padding.left    = typeof _.left   != 'undefined'  ? _.left    : padding.left;
    return this;
  };

  flow.margin =
  function( _ )
  {
    if ( !arguments.length ) return margin;
    flow.top     = typeof _.top    != 'undefined'  ? _.top     : flow.top;
    flow.right   = typeof _.right  != 'undefined'  ? _.right   : flow.right;
    flow.bottom  = typeof _.bottom != 'undefined'  ? _.bottom  : flow.bottom;
    flow.left    = typeof _.left   != 'undefined'  ? _.left    : flow.left;
    return this;
  };

  flow.width =
  function( value )
  {
    if ( !arguments.length ) return width;
    width = parseInt( value );
    return this;
  };

  flow.height =
  function( value )
  {
    if ( !arguments.length ) return height;
    height = parseInt( value );
    return this;
  };

  flow.nodes = 
  function( value )
  {
    if ( !arguments.length ) return nodes;
    nodes = value;
    return this;
  };

  flow.links = 
  function( value )
  {
    if ( !arguments.length ) return links;
    links = value;
    return this;
  };

  // TODO Unused functions - start
    flow.nodeWidth = function(value) {
        if (!arguments.length) return nodeWidth;
        nodeWidth = parseInt(value);
        return this;
    };

    flow.nodeHeight = function(value) {
        if (!arguments.length) return nodeHeight;
        nodeHeight = parseInt(value);
        return this;
    };

    flow.containerHeight = function(value) {
        if (!arguments.length) return containerHeight;
        containerHeight = parseInt(value);
        return this;
    };
  // TODO Unused functions - end 

  return flow;
}
