d3.custom.layout.flow = 
function() 
{
  var hierarchy = d3.layout.hierarchy().sort( null ).value( null ),
      width     = 0,
      height    = 0,
      padding   = { top:20, left:10, bottom:10, right:10 },
      margin    = { top:10, left:10, bottom:10, right:10 };

  var nodes = [],
//      oldNodes = [],
      links = [],
      visibleNodeIds = {},
      oldVisibleNodeIds = {};
  
  // TODO Unused variables
  var nodeWidth = 0,
      nodeHeight = 0,
      containerHeight = 0;
  
  // Replace all uses of nodeWidth and nodeHeight with node.iw and node.ih
  // containerWidth and containerHeight have to be calculated based on whether
  // the nodes is showing its children or not

  function 
  flow( d, i )
  {
    function
    gatherVisibleNodeIds( nodes )
    {
      // Remember old visible nodes 
//      oldVisibleNodeIds = visibleNodeIds;
      
      // Clear out the map
      visibleNodeIds = {};

      nodes.forEach( function( node ) {
        visibleNodeIds[ node.id2 ] = true;
      } );
    }

    // Returns an array of new nodes that need to be become visible
    // because they are/container the target of this node's edges
    function
    findNewNodesForLinks( node )
    {
      console.log( "Finding new links for node " + node.name )
      // Put the node in an array so that we can use "map"
      var nodes = [];
      var newNodes = [];
      nodes.push( node );
      // gather all links for visible nodes
      var newLinks = d3.merge( nodes.map( function( node ) {
        return ( node.edges || [] ).map( function( edge ) {
          // FIXME should we pass nodeIdMap to the constructor?
          return { source: nodeIdMap[ edge.from ], target: nodeIdMap[ edge.to ] };
        } )
      } ) );

      // For all new links, figure out which new nodes need to become
      // visible
      newLinks.forEach( function( link ) 
      {
        var toTurnOn = findNode( link.target );
        // TODO use merge?
        toTurnOn.forEach( function( d ) {
          newNodes.push( d );

          // Toggle the children for the node
          assert( d.children == null );
          assert( d._children != null );
          d.children = d._children;
          d._children = null;
        } );
      } );

      return newNodes;
    }
    function
    gatherLinks( nodes )
    {
      // gather all links for visible nodes
      var allLinks = d3.merge( nodes.map( function( node ) {
        return ( node.edges || [] ).map( function( edge ) {
          // FIXME should we pass nodeIdMap to the constructor?
          return { source: nodeIdMap[ edge.from ], target: nodeIdMap[ edge.to ] };
        } )
      } ) );

      // Go through all the links and only keep the ones where the target is
      // also visible 
      var visibleLinks = [];
      allLinks.forEach( function( link ) {
        // If the source of the link just became visible, also make the 
        // taget visible
//        if ( oldVisibleNodeIds[ link.source.id2 ] != true ) {
//          var toTurnOn = findNode( link.target );
//          toTurnOn.forEach( function( d ) {
//            assert( d.children == null );
//            assert( d._children != null );
//            d.children = d._children;
//            d._children = null;
//          } );
//        }
        
        // Otherwise, show the link only if the target is also visible
        if ( visibleNodeIds[ link.target.id2 ] == true )
          visibleLinks.push( link );
      } );

      return visibleLinks;
    }

    // If the node is visible, return an empty array 
    // If it's hidden it will return the array of nodes whose children need 
    // to be come visible for this node to be shown
    function
    findNode( node )
    {
      console.log( "Finding node: " + node.name );
      
      var allParents = [];
      if ( visibleNodeIds[ node.id2 ] == true ) return allParents;

      // TODO is parent a keyword?
      var nodeParent = node.parent;
      while ( nodeParent != null ) {
        // Children are visible
        if ( nodeParent.children ) break;
        
        console.log( "Adding parent " + nodeParent.name );
        allParents.push( nodeParent );
        nodeParent = nodeParent.parent;
      }
      
      return allParents;
    }

//    function
//    gatherLinks( nodes )
//    {
//      return d3.merge( nodes.map( function( node ) {
//        return ( node.edges || [] ).map( function( edge ) {
//          // FIXME should we pass nodeIdMap to the constructor?
//          return { source: nodeIdMap[ edge.from ], target: nodeIdMap[ edge.to ] };
//        } )
//      } ) );
//    }

    function
    firstWalk( node )
    {
      var children = node.children;
      // The node has visible children
      if ( children && children.length > 0 ) {
        children.forEach( firstWalk );
        layoutNode( node );
      }
      else {
        node.x = 0;
        node.y = 0;
        node.width = node.iw;
        node.height = node.ih;
      }
    }

    function
    secondWalk( node )
    {
      var children = node.children;
      if ( !children ) return;
      children.forEach( function( child ) {
        child.x += node.x;
        child.y += node.y;
        secondWalk( child );
      } );
    }

    // Layout a node and its children
    // This function should only be called on nodes that have (visible)
    // children
    function
    layoutNode( node )
    {
      // Lay out things differently based on the node type
      switch( node.type ) {
        case "memsys":
          layoutMemSystem( node );
          break;

        case "bank":
          layoutBank( node );
          break;

        default:
          gridLayout( node );
          break;
      }
    }

    function
    layoutMemSystem( node )
    {
      var   children  = node.children
          , curX      = 0     
          , curY      = 0     
          , width     = 0     
          ;

      children.forEach( function( child ) {
        child.x = curX;
        child.y = curY;
      
        curY += child.height;

        // Ensure all children have the same width
        if ( width == 0 ) width = child.width;
        else              assert( width == child.width );
      } );    

      node.width = width;
      node.height = curY;
      
      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }

    function
    layoutBank( node )
    {
      var   children  = node.children
          , curY      = padding.top/2 // TODO 
          ;
      
      children.forEach( function( child ) {
        child.x = 0;
        child.y = curY + child.height/2;

        curY += child.height/2 + padding.bottom;

        // Ensure that the child has a smaller width than this node
        assert( child.width < node.iw );
      } );

      node.width = node.iw;
      node.height = Math.max( curY, node.ih ); 

      node.x = 0;
      node.y = 0;
    }

    function
    gridLayout( node )
    {
      var   children  = node.children
          , curX      = 0 
          , curY      = 0 
          , maxWidth  = 0
          , maxHeight = 0
          , prev      = null;
      
      children.forEach( function( child ) {
        if ( prev ) {
          if ( prev.order == child.order ) {
            // same row, next column
            curX += prev.width + padding.right;
          }
          else {
            // next row, column 0
            curY += prev.height + padding.bottom;
            curX = 0;
          }
        }

        // (curX, curY) is the starting point of the cell
        // Move to the middle of the cell
        child.x = curX + padding.left;
        child.y = curY + padding.top;
        
        maxWidth = Math.max( maxWidth, child.x + child.width + padding.right );
        maxHeight = Math.max( maxHeight, child.y + child.height + padding.bottom );
        prev = child;
      } );

      node.width = maxWidth;
      node.height = maxHeight;

      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }

    // Layout a node and its children
    // This function should only be called on nodes that have (visible)
    // children
    function
    gridLayout0( node )
    {
      var   children  = node.children
          , curX      = padding.left
          , curY      = padding.top
          , maxWidth  = 0;
      
      children.forEach( function( child ) {
        child.x = curX;
        child.y = curY;

        curY += child.height + padding.bottom;
        maxWidth = Math.max( maxWidth, child.width + padding.right );
      } );

      node.width = curX + maxWidth;
      node.height = curY;
      node.x = node.parent ? padding.left : margin.left;
      node.y = node.parent ? padding.top  : margin.top;
    }
    
    oldVisibleNodeIds = visibleNodeIds;
    
    // Reset stuff
    nodes = hierarchy.call( this, d, i );
    var root  = nodes[ 0 ];
    links = [];
    
    console.log( "Initial flow nodes: " );
    console.log( nodes );

    // Figure out which of the nodes are newly added
    var newNodes = [];
    nodes.forEach( function( node )
    {
      if ( oldVisibleNodeIds[ node.id2 ] != true )
        newNodes.push( node );
    } );

    // For all newly added nodes, collect the edges
    // More nodes may have to be added based on what edges we want
    // to show, so nodes.length grows within the for loop
    for ( var i = 0; i < newNodes.length; ++i ) {
      var newNode = newNodes[ i ];
      var newNodesForLinks = findNewNodesForLinks( newNode );
      newNodesForLinks.forEach( function( d ) {
        newNodes.push( d );
      } );
    }
    
    // Now that we've turned some nodes, gather the nodes again
    nodes = hierarchy.call( this, d, i );
    
    console.log( "All flow nodes: " );
    console.log( nodes );

    // Now we know all visible nodes, remember their ids 
    gatherVisibleNodeIds( nodes );

    // Add all links for visible nodes. If a link's target is 
    // still not visible, then we don't add it
    links = gatherLinks( nodes );

    firstWalk( root );
    secondWalk( root );
    width = root.width;
    height = root.height; 

    return nodes;
  }

  // TODO understand what these are doing!
  flow.padding =
  function( _ )
  {
    if ( !arguments.length ) return padding;
    padding.top     = typeof _.top    != 'undefined'  ? _.top     : padding.top;
    padding.right   = typeof _.right  != 'undefined'  ? _.right   : padding.right;
    padding.bottom  = typeof _.bottom != 'undefined'  ? _.bottom  : padding.bottom;
    padding.left    = typeof _.left   != 'undefined'  ? _.left    : padding.left;
    return this;
  };

  flow.margin =
  function( _ )
  {
    if ( !arguments.length ) return margin;
    flow.top     = typeof _.top    != 'undefined'  ? _.top     : flow.top;
    flow.right   = typeof _.right  != 'undefined'  ? _.right   : flow.right;
    flow.bottom  = typeof _.bottom != 'undefined'  ? _.bottom  : flow.bottom;
    flow.left    = typeof _.left   != 'undefined'  ? _.left    : flow.left;
    return this;
  };

  flow.width =
  function( value )
  {
    if ( !arguments.length ) return width;
    width = parseInt( value );
    return this;
  };

  flow.height =
  function( value )
  {
    if ( !arguments.length ) return height;
    height = parseInt( value );
    return this;
  };

  flow.nodes = 
  function( value )
  {
    if ( !arguments.length ) return nodes;
    nodes = value;
    return this;
  };

  flow.links = 
  function( value )
  {
    if ( !arguments.length ) return links;
    links = value;
    return this;
  };

  // TODO Unused functions - start
    flow.nodeWidth = function(value) {
        if (!arguments.length) return nodeWidth;
        nodeWidth = parseInt(value);
        return this;
    };

    flow.nodeHeight = function(value) {
        if (!arguments.length) return nodeHeight;
        nodeHeight = parseInt(value);
        return this;
    };

    flow.containerHeight = function(value) {
        if (!arguments.length) return containerHeight;
        containerHeight = parseInt(value);
        return this;
    };
  // TODO Unused functions - end 

  return flow;
}
