// Define size of separator (in pixels) relative to window size.
var SEPARATOR_SIZE = Math.floor(window.innerWidth*0.25/100) + 1;

/* 
    Define classes.
    3 types: 
        GroupPanels represent divs which hold other panels(possibly more groups) and separators.
            They divide either horizontally or vertically.
        SinglePanels represent divs that have no children, but have ids specified by LAYOUT_SETTINGS.
        Separators are placed in GroupPanels' children property in between every SinglePanel.  
*/

function GroupPanel(caste){
    this.caste = caste;
    this.separator = (caste == "groupHorizontal")? "width": "height";
    this.width = "100%";
    this.height = "100%";
	this.children = [];
}

function SinglePanel(id){
    this.id = id;
    this.caste = "singlePanel";
    this.width = "100%";
    this.height = "100%";    
}

function Separator(){
    this.caste = "separator";
    this.width = "100%";
    this.height = "100%";    
}




/*
    Recursively create the HTML layout.
*/

function paintLayout(currentPanel, currentDiv){
	if (currentPanel instanceof GroupPanel){
		// Update
		var children = currentDiv.selectAll("div")
						.data(currentPanel.children);
		
		// Enter
		children.enter().append("div");

		children.attr("data-caste", function (d){return d.caste;})
			.style("height", function (d){return d.height;})
			.style("width", function (d){return d.width;})
			.style("float", "none");
		
		children.style(currentPanel.separator, function (d) {
			var total = parseInt(currentDiv.style(currentPanel.separator)),
				current = parseInt(d3.select(this).style(currentPanel.separator)),
				remove = (currentPanel.children.length-1)/(currentPanel.children.length+1) * SEPARATOR_SIZE;
			return ((current-remove)/total*100)+"%";
		});
		
		if (currentPanel.separator == "width"){
			children.style("float", "left"); 
		}
	
		children.filter('[data-caste="singlePanel"]')
			.attr("id", function (d){return d.id;});
//            .call(swap);

		children.filter('[data-caste="separator"]')
			.style(currentPanel.separator , ((SEPARATOR_SIZE/parseInt(currentDiv.style(currentPanel.separator))*100)+"%"))
			.call(drag);
		
		//loader.print(currentPanel.separator + " " + SEPARATOR_SIZE + " " + parseInt(currentDiv.style(currentPanel.separator)));
		children.each(function (d) { paintLayout(d, d3.select(this));});
		
		children.exit().remove();
		
	}
}


/*
    Define dragging behaviour for panel switching.
*/

var swap = d3.behavior.drag()
        .on("dragstart", initSwap)
        .on("drag", showSwap)
        .on("dragend", finishSwap);
        
var swapWhat, swapWith, prevSib;

function initSwap(){
    swapWhat = chaseSelector(this, '[data-caste="singlePanel"]');
    prevSib = document.createElement("div");
    swapWhat.parentNode.insertBefore(prevSib, swapWhat)
    swapWith = swapWhat;    

}

function finishSwap(){
    removeNode(prevSib);
}

function showSwap(){
    var elementOver = document.elementFromPoint(d3.mouse(document.body)[0], d3.mouse(document.body)[1]);
    if (isInsideSelector(elementOver, "#layout") && 
        isInsideSelector(elementOver, '[data-caste="singlePanel"]') &&
        !isInside(elementOver, swapWhat)){
        if (swapWith != chaseSelector(elementOver, '[data-caste="singlePanel"]')) doSwap();
        swapWith = chaseSelector(elementOver, '[data-caste="singlePanel"]');
        doSwap();
        if (swapWhat.previousSibling == prevSib) swapWith = swapWhat;
    }
}   

function doSwap(){
    var panel1 = swapWhat,
        panel2 = swapWith;    

    while (panel1 != document.getElementById("layout") && !(d3.select(panel1).attr("data-caste") == "singlePanel"))
        panel1 = panel1.parentNode;

    while (panel2 != document.getElementById("layout") && !(d3.select(panel2).attr("data-caste") == "singlePanel"))
        panel2 = panel2.parentNode;

    if (d3.select(panel1).attr("data-caste") == "singlePanel" 
       && d3.select(panel2).attr("data-caste") == "singlePanel") {
         swapElements(panel1, panel2);
		 //panel1.dispatchEvent(changed);
		 //panel2.dispatchEvent(changed);
    }  
}



/*
    Define dragging behaviour of the separators. When dragged,
        either the height or width (depending on type of split) of the
        next and previous DOM elements (which the separator divides) 
        is adjusted accordingly.  
*/

var drag = d3.behavior.drag()
	.origin(function(d) { return {x: this.offsetLeft, y: this.offsetTop}; })
    .on("dragstart", draginit)
    .on("drag", dragmove);

var change = {},
    dimension = "",
    value = 0,
	move = 0,
    prev,
	prevSize,
    next,
	nextSize;

function draginit(d){
    dimension = d3.select(this.parentNode).datum().separator;
    value = {width:this.parentNode.offsetWidth, height:this.parentNode.offsetHeight}[dimension];
	change = {width:0, height:0};  
    prev = d3.select(this.previousSibling);
	prevSize = parseInt(prev.style(dimension));
    next = d3.select(this.nextSibling);
	nextSize = parseInt(next.style(dimension));
	move = 0;
	//loader.print(prev.attr("id") + " " + next.attr("id"));
}
	
function dragmove(d) {
    change.width += d3.event.dx;
    change.height += d3.event.dy;
    var move = change[dimension];
    if (prevSize + move < 0) 
        move = -prevSize;
    if (nextSize - move < 0)
        move = nextSize;
    prev.style(dimension, ((prevSize + move)*100/value)+"%");
    next.style(dimension, ((nextSize - move)*100/value)+"%");
	//prev[0][0].dispatchEvent(changed);
	//next[0][0].dispatchEvent(changed);

}

/*
    Reads in an array of settings and creates the specified hierarchical structure, 
        returning a root GroupPanel object.
*/

function initLayout(layout){
    var stack = [],
        stackTop = 0,
        count = 0,
        curr;
    var mainPanel = new GroupPanel("");
    stack[stackTop] = mainPanel;

    while (count < layout.length) {
        switch(layout[count]) {
            case "$H(":
                curr = new GroupPanel("groupHorizontal");
                addPanel(curr, stack[stackTop++]);
                stack.push(curr);
                break;
            case "$V(":
                curr = new GroupPanel("groupVertical");
                addPanel(curr, stack[stackTop++]);
                stack.push(curr);
                break;
            case ")":
                curr = stack[stackTop--];
                stack.pop();
                break;
            case "%":
                count++;
                curr[stack[stackTop].separator] = layout[count] + "%";
                break;
            default:
                curr = new SinglePanel(layout[count]);
                addPanel(curr, stack[stackTop]);

        }
        count++;
    }
    return mainPanel;
}

// Helper function. Add child panel to parent panel. Insert separator as necessary. 
function addPanel(newPanel, group){
    if (group.children.length){
        group.children.push(new Separator());    
    }    
    group.children.push(newPanel);
}


/*
    Main function. Call it with settings argument to create layout.
    The settings variable is a string as such:
        "$V(" + <child panels> + ")" defines a GroupPanel with the height dimension distributed across children.
        "$H(" + <child panels> + ")" defines a GroupPanel with the width dimension distributed across children.
        A single Panel is represented simply by the desired id of the panel.  
        Every child panel (group or single) is followed by a "% x" to define the percentage (x) of the parent 
            element's divided dimension taken up by the panel.
*/

function Layout(settings){
	var mainPanel = initLayout(settings.split(" "));
	d3.select("#layout").attr("id", "temp");
	
	var canvas = d3.select("body")
					.append("div")
					.attr("id", "layout")
					.style("width", (window.innerWidth - 20)+"px")
					.style("height", (window.innerHeight - 20)+"px");

	paintLayout(mainPanel, canvas);

	d3.select("#layout").selectAll('div[data-caste="singlePanel"]')
		.each( function(d) {
			var x = d3.select("#temp").selectAll('div[data-caste="singlePanel"]').filter("#"+d.id);
			if (x.size()) {swapElements(x[0][0], this);}
		}	
		);
	
	d3.select("#temp").each(
		function(){ 
			removeNode(this); 
		}
	);
	/* d3.select("body").selectAll("div")
		.each(function(){
			loader.print(d3.select(this).attr("id") + "::" + d3.select(this).attr("data-caste") + " ---> (" + d3.select(this).style("width") + " , " + d3.select(this).style("height"));
		});
	 */	
}


/* 
    Define the facility to move Panels (and DOM Nodes in general)
    around the layout, as well as other useful functions.
*/

function moveNode(element, newParent, newNext){
    if (newNext) {
        newParent.insertBefore(element, newNext);
    } else {
        newParent.appendChild(element);
    }
}

function replaceNode(newElement, oldElement){
    oldElement.parentNode.insertBefore(newElement, oldElement);
    removeNode(oldElement);
}

function removeNode(element){
	element.parentNode.removeChild(element);
}

function isInside(element, container){
    return (!element)? false: ((element == container)? true: isInside(element.parentNode, container));
    
}

function isInsideSelector(element, containerSelector){
    return (!element)? false: ((d3.select(element).filter(containerSelector).size())? 
                                true: isInsideSelector(element.parentNode, containerSelector));
}
function chaseSelector(element, selector){
    return (!element)? null :(d3.select(element).filter(selector).size()?element:chaseSelector(element.parentNode, selector))    
}

function swapProperty(element1, element2, property){
    var x = d3.select(element1),
        y = d3.select(element2);
		
    var temp = x.style(property);
    x.style(property, y.style(property));
    y.style(property, temp);
}

function swapDimension(element1, element2, property){
    var x = d3.select(element1),
		pX = d3.select(element1.parentNode),
        y = d3.select(element2),
		pY = d3.select(element2.parentNode);	
	var temp = (parseInt(x.style(property))/parseInt(pX.style(property)) * 100) + "%";
    x.style(property, (parseInt(y.style(property))/parseInt(pY.style(property)) * 100) + "%");
    y.style(property, temp);
}

function swapElements(element1, element2){
    // Swap environment properties
    swapDimension(element1, element2, "width");
    swapDimension(element1, element2, "height");
    swapProperty(element1, element2, "float");

    // Swap positions
    var parent1 = element1.parentNode,
        parent2 = element2.parentNode,
        next1 = element1.nextSibling,
        next2 = element2.nextSibling;
    moveNode(element1, parent2, next2);
    moveNode(element2, parent1, next1);
}

function getStructure(element, depth){
	var str = (Array(depth*4 + 1).join(" ")) + (element.hasAttribute("id")? element.getAttribute("id"):"") + " " + (element.hasAttribute("data-caste")? element.getAttribute("data-caste"):"") + "\n"
	console.log(str);
	var child = element.firstChild;
	while (child){
		str += getStructure(child, depth+1);
		child = child.nextSibling;
	}
}
