# (C) 1992-2016 Altera Corporation. All rights reserved.                         
# Your use of Altera Corporation's design tools, logic functions and other       
# software and tools, and its AMPP partner logic functions, and any output       
# files any of the foregoing (including device programming or simulation         
# files), and any associated documentation or information are expressly subject  
# to the terms and conditions of the Altera Program License Subscription         
# Agreement, Altera MegaCore Function License Agreement, or other applicable     
# license agreement, including, without limitation, that your use is for the     
# sole purpose of programming logic devices manufactured by Altera and sold by   
# Altera or its authorized distributors.  Please refer to the applicable         
# agreement for further details.                                                 
    


# This TCL script creates a qsys simulation system which includes all the hls generated components
# The qsys system also includes the main DPI controller and per-component DPI cosimulation drivers

package require -exact qsys 14.0

# "sim_qsys" and "component_list" variables have already been set by a previous TCL command

# -----------------------------------------------------------------------------
# Debugging
# -----------------------------------------------------------------------------
set DEBUG_MSGS 0

# -----------------------------------------------------------------------------
# Define procedures used to instantiate and connect the DPI controllers/drivers
# -----------------------------------------------------------------------------

proc debug { s } {
    global DEBUG_MSGS
    if { $DEBUG_MSGS } {
        puts $s
    }
}

proc assert {cond {msg "assertion failed"}} {
    if {![uplevel 1 expr $cond]} {error $msg}
}

proc instantiate_dpi_bfm_and_connect_to_implicit_stream_sink { component interface } {

    add_instance stream_source_dpi_bfm_${component}_${interface}_inst hls_sim_stream_source_dpi_bfm
    set_instance_parameter_value stream_source_dpi_bfm_${component}_${interface}_inst FORCE_STREAM_CONDUIT 1
    
    add_connection clock_reset_inst.clock      stream_source_dpi_bfm_${component}_${interface}_inst.clock
    add_connection clock_reset_inst.clock2x    stream_source_dpi_bfm_${component}_${interface}_inst.clock2x
    add_connection clock_reset_inst.reset      stream_source_dpi_bfm_${component}_${interface}_inst.reset
    
    set stream_width [get_instance_interface_port_property ${component}_inst ${interface} ${interface} WIDTH]
    set_instance_parameter_value stream_source_dpi_bfm_${component}_${interface}_inst STREAM_DATAWIDTH $stream_width

    set_instance_parameter_value stream_source_dpi_bfm_${component}_${interface}_inst INTERFACE_NAME $interface
    set_instance_parameter_value stream_source_dpi_bfm_${component}_${interface}_inst COMPONENT_NAME $component

    add_connection stream_source_dpi_bfm_${component}_${interface}_inst.source_data ${component}_inst.${interface}
   
}

proc instantiate_dpi_bfm_and_connect_to_stream_sink { component interface } {

    add_instance stream_source_dpi_bfm_${component}_${interface}_inst hls_sim_stream_source_dpi_bfm
    
    add_connection clock_reset_inst.clock      stream_source_dpi_bfm_${component}_${interface}_inst.clock
    add_connection clock_reset_inst.clock2x    stream_source_dpi_bfm_${component}_${interface}_inst.clock2x
    add_connection clock_reset_inst.reset      stream_source_dpi_bfm_${component}_${interface}_inst.reset
    
    set stream_width [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_data WIDTH]
    set_instance_parameter_value stream_source_dpi_bfm_${component}_${interface}_inst STREAM_DATAWIDTH $stream_width

    set_instance_parameter_value stream_source_dpi_bfm_${component}_${interface}_inst INTERFACE_NAME $interface
    set_instance_parameter_value stream_source_dpi_bfm_${component}_${interface}_inst COMPONENT_NAME $component

    add_connection stream_source_dpi_bfm_${component}_${interface}_inst.source ${component}_inst.${interface}
   
}


proc instantiate_dpi_bfm_and_connect_to_stream_source { component interface } {

    add_instance stream_sink_dpi_bfm_${component}_${interface}_inst hls_sim_stream_sink_dpi_bfm
    
    add_connection clock_reset_inst.clock      stream_sink_dpi_bfm_${component}_${interface}_inst.clock
    add_connection clock_reset_inst.clock2x    stream_sink_dpi_bfm_${component}_${interface}_inst.clock2x
    add_connection clock_reset_inst.reset      stream_sink_dpi_bfm_${component}_${interface}_inst.reset
    
    set stream_width [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_data WIDTH]
    set_instance_parameter_value stream_sink_dpi_bfm_${component}_${interface}_inst STREAM_DATAWIDTH $stream_width

    set_instance_parameter_value stream_sink_dpi_bfm_${component}_${interface}_inst INTERFACE_NAME $interface
    set_instance_parameter_value stream_sink_dpi_bfm_${component}_${interface}_inst COMPONENT_NAME $component

    add_connection ${component}_inst.${interface} stream_sink_dpi_bfm_${component}_${interface}_inst.sink

}


proc instantiate_dpi_bfm_and_connect_to_avalon_master { component interface } {

    #assert [regexp {(avm_memgmem0_(\d+)_port_.*)} $interface full tmp id] "Unexpected name for Avalon master interface"
    regexp {avm_memgmem[0-9]+_([0-9]+)_.*} $interface full id

    set dpi_bfm_inst mm_slave_dpi_bfm_${component}_${interface}_inst
    add_instance $dpi_bfm_inst hls_sim_mm_slave_dpi_bfm
    
    set_instance_parameter_value $dpi_bfm_inst INTERFACE_ID   [expr $id]
    set_instance_parameter_value $dpi_bfm_inst COMPONENT_NAME $component

    set AV_ADDRESS_W [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_address WIDTH]
    set_instance_parameter_value $dpi_bfm_inst AV_ADDRESS_W $AV_ADDRESS_W

    # symbol width in bits, symbol is one byte
    set AV_SYMBOL_W 8

    set AV_READDATA_W [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_readdata WIDTH]
    set AV_NUMSYMBOLS [ expr {$AV_READDATA_W / $AV_SYMBOL_W}]
    set AV_NUMSYMBOLS_REM [ expr {$AV_READDATA_W % $AV_SYMBOL_W}]
    assert {$AV_NUMSYMBOLS>0 && $AV_NUMSYMBOLS_REM==0} "AV_READDATA_W must be a multiple of AV_SYMBOL_W=$AV_SYMBOL_W"
    set_instance_parameter_value $dpi_bfm_inst AV_NUMSYMBOLS $AV_NUMSYMBOLS
    set_instance_parameter_value $dpi_bfm_inst AV_SYMBOL_W $AV_SYMBOL_W
    
    set AV_BYTEENABLE_W [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_byteenable WIDTH]
    assert {$AV_NUMSYMBOLS==$AV_BYTEENABLE_W} "AV_NUMSYMBOLS=$AV_NUMSYMBOLS must be equal to the width of the byteenable signal ($AV_BYTEENABLE_W)"

    set AV_FIX_READ_LATENCY [get_instance_interface_parameter_value ${component}_inst $interface readLatency]
    set_instance_parameter_value $dpi_bfm_inst AV_FIX_READ_LATENCY $AV_FIX_READ_LATENCY

    set ADDRESS_UNITS [get_instance_interface_parameter_value ${component}_inst $interface addressUnits]
    set_instance_parameter_value $dpi_bfm_inst ADDRESS_UNITS $ADDRESS_UNITS

    set_instance_parameter_value $dpi_bfm_inst USE_READ_DATA_VALID 1

    # burst count width
    set AV_BURSTCOUNT_W [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_burstcount WIDTH]
    set_instance_parameter_value $dpi_bfm_inst AV_BURSTCOUNT_W $AV_BURSTCOUNT_W

    if { $AV_BURSTCOUNT_W > 1 } {
        set_instance_parameter_value $dpi_bfm_inst USE_BURSTCOUNT 1
    }

    add_connection clock_reset_inst.clock      $dpi_bfm_inst.clock    
    add_connection clock_reset_inst.reset      $dpi_bfm_inst.reset
    add_connection ${component}_inst.${interface} $dpi_bfm_inst.s0

}

proc instantiate_dpi_bfm_and_connect_to_avalon_slave { component interface } {

    set dpi_bfm_inst mm_master_dpi_bfm_${component}_${interface}_inst
    add_instance $dpi_bfm_inst hls_sim_mm_master_dpi_bfm
    
    set_instance_parameter_value $dpi_bfm_inst COMPONENT_NAME $component

    #set AV_ADDRESS_W [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_address WIDTH]
    set AV_ADDRESS_W 64
    set_instance_parameter_value $dpi_bfm_inst AV_ADDRESS_W $AV_ADDRESS_W

    # symbol width in bits, symbol is one byte
    set AV_SYMBOL_W 8

    #set AV_READDATA_W [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_readdata WIDTH]
    set AV_READDATA_W 64
    set AV_NUMSYMBOLS [ expr {$AV_READDATA_W / $AV_SYMBOL_W}]
    set AV_NUMSYMBOLS_REM [ expr {$AV_READDATA_W % $AV_SYMBOL_W}]
    assert {$AV_NUMSYMBOLS>0 && $AV_NUMSYMBOLS_REM==0} "AV_READDATA_W must be a multiple of AV_SYMBOL_W=$AV_SYMBOL_W"
    set_instance_parameter_value $dpi_bfm_inst AV_NUMSYMBOLS $AV_NUMSYMBOLS
    set_instance_parameter_value $dpi_bfm_inst AV_SYMBOL_W $AV_SYMBOL_W
    
    set AV_BYTEENABLE_W [get_instance_interface_port_property ${component}_inst ${interface} ${interface}_byteenable WIDTH]
    assert {$AV_NUMSYMBOLS==$AV_BYTEENABLE_W} "AV_NUMSYMBOLS=$AV_NUMSYMBOLS must be equal to the width of the byteenable signal ($AV_BYTEENABLE_W)"

    set AV_FIX_READ_LATENCY [get_instance_interface_parameter_value ${component}_inst $interface readLatency]
    set_instance_parameter_value $dpi_bfm_inst AV_FIX_READ_LATENCY $AV_FIX_READ_LATENCY

    set_instance_parameter_value $dpi_bfm_inst USE_READ_DATA_VALID 1

    # burst count width
    set AV_BURSTCOUNT_W 1

    add_connection clock_reset_inst.clock      $dpi_bfm_inst.clock    
    add_connection clock_reset_inst.reset      $dpi_bfm_inst.reset
    add_connection $dpi_bfm_inst.m0 ${component}_inst.${interface}

}

proc add_dpi_controller_for_component { component num_iface comp_has_return_iface num_implicit_iface comp_has_slave_iface } {

    # Add component_dpi_controller instance  (one per component so component name must be in the inst name)
    add_instance component_dpi_controller_${component}_inst hls_sim_component_dpi_controller

    # Connect clock/reset to the component_dpi_controller instance
    add_connection clock_reset_inst.clock       component_dpi_controller_${component}_inst.clock
    add_connection clock_reset_inst.clock2x     component_dpi_controller_${component}_inst.clock2x
    add_connection clock_reset_inst.reset       component_dpi_controller_${component}_inst.reset

    # Connect implicit input/output streams between the component_dpi_controller and the component
    add_connection component_dpi_controller_${component}_inst.component_call ${component}_inst.call

    if { $comp_has_return_iface == 1 } {
        add_connection ${component}_inst.return component_dpi_controller_${component}_inst.component_return

        if { [lsearch -exact [get_instance_interfaces ${component}_inst] returndata] >= 0 } {
          set stream_width [get_instance_interface_port_property ${component}_inst returndata returndata WIDTH]
          set_instance_parameter_value component_dpi_controller_${component}_inst RETURN_DATAWIDTH $stream_width
          add_connection ${component}_inst.returndata component_dpi_controller_${component}_inst.returndata
        }
    } 

    # must set COMPONENT_NAME parameter of the component_dpi_controller
    # set it to the original component name ($component) not ${component}_inst
    set_instance_parameter_value component_dpi_controller_${component}_inst COMPONENT_NAME $component 
    
    # add Avalon Conduit Fanout module for go_bind signal
    add_instance ${component}_component_dpi_controller_bind_conduit_fanout_inst avalon_conduit_fanout
    set_instance_parameter_value ${component}_component_dpi_controller_bind_conduit_fanout_inst numFanOut $num_iface
    
    # add Avalon Conduit Fanout module for enable signal
    add_instance ${component}_component_dpi_controller_enable_conduit_fanout_inst avalon_conduit_fanout
    set_instance_parameter_value ${component}_component_dpi_controller_enable_conduit_fanout_inst numFanOut $num_iface

    # add Avalon Conduit Fanout module for implicit stream ready signals
    if { $num_implicit_iface > 0 } {
        add_instance ${component}_component_dpi_controller_implicit_ready_conduit_fanout_inst avalon_conduit_fanout
        set_instance_parameter_value ${component}_component_dpi_controller_implicit_ready_conduit_fanout_inst numFanOut $num_implicit_iface

        add_connection component_dpi_controller_${component}_inst.read_implicit_streams  ${component}_component_dpi_controller_implicit_ready_conduit_fanout_inst.in_conduit
    }

    add_connection component_dpi_controller_${component}_inst.dpi_control_bind  ${component}_component_dpi_controller_bind_conduit_fanout_inst.in_conduit
    add_connection component_dpi_controller_${component}_inst.dpi_control_enable  ${component}_component_dpi_controller_enable_conduit_fanout_inst.in_conduit

    # there is only ever one Avalon Master BFM for the component's slave, so just connect the slave ready signal
    if { $comp_has_slave_iface == 1 } {
      set_instance_parameter_value component_dpi_controller_${component}_inst WAIT_FOR_SLAVE 1
      add_connection mm_master_dpi_bfm_${component}_avs_cra_inst.dpi_control_done_writes component_dpi_controller_${component}_inst.dpi_control_slave_ready
    }

}

proc need_dpi_bfm_inst_for_interface { component interface } {
    # we already connected these
    if { $interface != "call" && $interface != "return" && $interface != "returndata" && $interface != "clock" &&
         $interface != "clock2x" && $interface != "reset"  } {
        return 1
    } else {
        return 0
    }            
}

proc instantiate_component_and_all_its_dpi_modules { component } {
    add_instance ${component}_inst $component

    # Connect clock/reset to the component 
    add_connection clock_reset_inst.clock       ${component}_inst.clock
    add_connection clock_reset_inst.reset       ${component}_inst.reset

    # Walk through component interfaces and instantiate correct dpi_bfm for each 
    # Do not connect implict interfaces and clock/reset (we already connected them)
    # Also, if the component has clock2x (optional) then connect it
    set num_iface 0
    set num_implicit_iface 0
    set comp_has_return_iface 0
    set comp_has_slave_iface 0
    foreach interface [get_instance_interfaces ${component}_inst] {     
        set description [get_instance_interface_property ${component}_inst $interface DESCRIPTION]
        debug $description
        set class_name [get_instance_interface_property ${component}_inst $interface CLASS_NAME]
        debug $class_name
        set portlist [get_instance_interface_ports ${component}_inst $interface]
        set numports [llength $portlist]
        if {$interface == "clock2x"} {
            add_connection clock_reset_inst.clock2x     ${component}_inst.clock2x
        } elseif {$interface == "return"} {
            set comp_has_return_iface 1
        } elseif { [need_dpi_bfm_inst_for_interface $component $interface] } {
            if { $class_name=="avalon_streaming_sink" || ($class_name=="conduit_end" && $numports==3) } {
                instantiate_dpi_bfm_and_connect_to_stream_sink $component $interface
                set  num_iface [expr $num_iface + 1]
            } elseif { $class_name=="avalon_streaming_source" || ($class_name=="conduit_start" && $numports==3) } {
                instantiate_dpi_bfm_and_connect_to_stream_source $component $interface
                set  num_iface [expr $num_iface + 1]
            } elseif { $class_name=="avalon_master"} {
                instantiate_dpi_bfm_and_connect_to_avalon_master $component $interface
                set  num_iface [expr $num_iface + 1]
            } elseif { $class_name=="avalon_slave"} {
                instantiate_dpi_bfm_and_connect_to_avalon_slave $component $interface
                set  comp_has_slave_iface 1
                set  num_iface [expr $num_iface + 1]
                set  num_implicit_iface [expr $num_implicit_iface + 1]
            } elseif { $class_name == "conduit_end" && $numports == 1} {
                instantiate_dpi_bfm_and_connect_to_implicit_stream_sink $component $interface
                set  num_iface [expr $num_iface + 1]
                set  num_implicit_iface [expr $num_implicit_iface + 1]
            } else {
                send_message error "Unrecognized interface class on $component.$interface: $class_name\n";
            }
        }
    }

    add_dpi_controller_for_component $component $num_iface $comp_has_return_iface $num_implicit_iface $comp_has_slave_iface

    # Add control signals from the component DPI controller to the interface BFMs
    set num_iface 0
    set num_implicit_iface 0
    foreach interface [get_instance_interfaces ${component}_inst] {     
        set portlist [get_instance_interface_ports ${component}_inst $interface]
        set numports [llength $portlist]
        if { [need_dpi_bfm_inst_for_interface $component $interface] } {
            set class_name [get_instance_interface_property ${component}_inst $interface CLASS_NAME]
            if { $class_name=="avalon_streaming_sink" || ($class_name=="conduit_end" && $numports==3) } {
                add_connection ${component}_component_dpi_controller_bind_conduit_fanout_inst.out_conduit_$num_iface   stream_source_dpi_bfm_${component}_${interface}_inst.dpi_control_bind
                add_connection ${component}_component_dpi_controller_enable_conduit_fanout_inst.out_conduit_$num_iface stream_source_dpi_bfm_${component}_${interface}_inst.dpi_control_enable
                set  num_iface [expr $num_iface + 1]
            } elseif { $class_name=="avalon_streaming_source" || ($class_name=="conduit_start" && $numports==3) } {
                add_connection ${component}_component_dpi_controller_bind_conduit_fanout_inst.out_conduit_$num_iface   stream_sink_dpi_bfm_${component}_${interface}_inst.dpi_control_bind
                add_connection ${component}_component_dpi_controller_enable_conduit_fanout_inst.out_conduit_$num_iface stream_sink_dpi_bfm_${component}_${interface}_inst.dpi_control_enable
                set  num_iface [expr $num_iface + 1]
            } elseif { $class_name=="avalon_master"} {
                add_connection ${component}_component_dpi_controller_bind_conduit_fanout_inst.out_conduit_$num_iface   mm_slave_dpi_bfm_${component}_${interface}_inst.dpi_control_bind
                add_connection ${component}_component_dpi_controller_enable_conduit_fanout_inst.out_conduit_$num_iface mm_slave_dpi_bfm_${component}_${interface}_inst.dpi_control_enable
                set  num_iface [expr $num_iface + 1]
            } elseif { $class_name=="avalon_slave"} {
                add_connection ${component}_component_dpi_controller_bind_conduit_fanout_inst.out_conduit_$num_iface   mm_master_dpi_bfm_${component}_${interface}_inst.dpi_control_bind
                add_connection ${component}_component_dpi_controller_enable_conduit_fanout_inst.out_conduit_$num_iface mm_master_dpi_bfm_${component}_${interface}_inst.dpi_control_enable
                add_connection ${component}_component_dpi_controller_implicit_ready_conduit_fanout_inst.out_conduit_$num_implicit_iface mm_master_dpi_bfm_${component}_${interface}_inst.dpi_control_component_started
                set  num_iface [expr $num_iface + 1]
                set  num_implicit_iface [expr $num_implicit_iface + 1]
            } elseif { $class_name == "conduit_end" && $numports == 1} {
                add_connection ${component}_component_dpi_controller_bind_conduit_fanout_inst.out_conduit_$num_iface   stream_source_dpi_bfm_${component}_${interface}_inst.dpi_control_bind
                add_connection ${component}_component_dpi_controller_enable_conduit_fanout_inst.out_conduit_$num_iface stream_source_dpi_bfm_${component}_${interface}_inst.dpi_control_enable
                add_connection ${component}_component_dpi_controller_implicit_ready_conduit_fanout_inst.out_conduit_$num_implicit_iface stream_source_dpi_bfm_${component}_${interface}_inst.source_ready
                set  num_iface [expr $num_iface + 1]
                set  num_implicit_iface [expr $num_implicit_iface + 1]
            } else {
                send_message error "Unrecognized interface class on $component.$interface: $class_name\n";
            }
        }
    }
}


# -----------------------------------------------------------------------------
# Create a new qsys system
# -----------------------------------------------------------------------------
append sim_qsys "_sim"
create_system $sim_qsys
puts "Generating qsys simulation system: $sim_qsys"

# Create a TCL list from a comma-separated string
set components [split $component_list ","]

# Add the clock/reset instance, one per sim system
add_instance clock_reset_inst hls_sim_clock_reset

# Add main_dpi_controller_inst, one per sim system
add_instance main_dpi_controller_inst hls_sim_main_dpi_controller
set_instance_parameter_value main_dpi_controller_inst NUM_COMPONENTS [llength $components]
set_instance_parameter_value main_dpi_controller_inst COMPONENT_NAMES_STR $component_list

# Connect clock/reset to the main_dpi_controller instance
add_connection clock_reset_inst.clock              main_dpi_controller_inst.clock
add_connection clock_reset_inst.clock2x            main_dpi_controller_inst.clock2x
add_connection clock_reset_inst.reset              main_dpi_controller_inst.reset
add_connection main_dpi_controller_inst.reset_ctrl clock_reset_inst.reset_ctrl

# -----------------------------------------------------------------------------
# For each component, instantiate and connect the controller and DPI interface drivers 
# -----------------------------------------------------------------------------
foreach component $components {     
    instantiate_component_and_all_its_dpi_modules $component
}

# -----------------------------------------------------------------------------
# Connect main_dpi_controller to all component_dpi_controllers (one per component)
# This is done using avalon_split_multibit_conduit and avalon_concatenate_singlebit_conduits adapters
# -----------------------------------------------------------------------------

add_instance split_component_start_inst avalon_split_multibit_conduit
set_instance_parameter_value split_component_start_inst multibit_width [llength $components]

add_connection main_dpi_controller_inst.component_enabled split_component_start_inst.in_conduit
set i 0
foreach component $components {     
    add_connection split_component_start_inst.out_conduit_$i component_dpi_controller_${component}_inst.component_enabled
    incr i
}

add_instance concatenete_component_done_inst avalon_concatenate_singlebit_conduits
set_instance_parameter_value concatenete_component_done_inst multibit_width [llength $components]

add_connection concatenete_component_done_inst.out_conduit main_dpi_controller_inst.component_done
set i 0
foreach component $components {     
    add_connection component_dpi_controller_${component}_inst.component_done concatenete_component_done_inst.in_conduit_$i 
    incr i
}

# Create the qsys system
save_system $sim_qsys

